from UnionFind import UnionFind
import copy
import unittest


class TestUnionFind(unittest.TestCase):

    uf = UnionFind()
    uf.get_or_insert(1)
    uf.get_or_insert(10)
    uf.get_or_insert(2)
    uf.get_or_insert(5)

    def test_insert_or_get(self):
        uf = copy.deepcopy(self.uf)
        self.assertIn(1, uf)
        self.assertIn(10, uf)
        self.assertIn(2, uf)
        self.assertIn(5, uf)
        self.assertEqual(uf.get_or_insert(1), 1)
        self.assertEqual(uf.get_or_insert(5), 5)
        self.assertEqual(uf.get_or_insert(10), 10)
        self.assertEqual(uf.get_or_insert(2), 2)
        self.assertEqual(uf.get_or_insert(30), 30)
        self.assertEqual(uf.get_or_insert(52), 52)

    def test_get(self):
        uf = self.uf
        self.assertEqual(uf[1], 1)
        self.assertEqual(uf[10], 10)

    def test_get_error(self):
        with self.assertRaises(Exception):
            _ = self.uf[52]
        with self.assertRaises(Exception):
            _ = self.uf[30]

    def test_union(self):
        uf = copy.deepcopy(self.uf)
        uf.union(1, 10)
        self.assertEqual(uf.get_or_insert(1), uf.get_or_insert(10))
        uf.union(2, 5)
        self.assertEqual(uf.get_or_insert(2), uf.get_or_insert(5))
        uf.union(2, 1)
        self.assertEqual(uf.get_or_insert(10), uf.get_or_insert(5))



"""
Functions (unary and binary) for use in Raco.
"""

import math
import md5

from .expression import (ZeroaryOperator, UnaryOperator, BinaryOperator,
                         NaryOperator, types, check_is_numeric, check_type,
                         TypeSafetyViolation)


class UnaryFunction(UnaryOperator):
    def __str__(self):
        return "%s(%s)" % (self.__class__.__name__, self.input)

    def __repr__(self):
        return "{op}({inp!r})".format(op=self.opname(), inp=self.input)


class BinaryFunction(BinaryOperator):
    def __str__(self):
        return "%s(%s, %s)" % (self.__class__.__name__, self.left, self.right)

    def __repr__(self):
        return "{op}({l!r}, {r!r})".format(op=self.opname(), l=self.left,
                                           r=self.right)


class NaryFunction(NaryOperator):
    def __str__(self):
        return "%s(%s)" % \
            (self.__class__.__name__,
             ",".join([str(op) for op in self.operands]))

    def __repr__(self):
        return "{op}({ch!r})".format(op=self.opname(), ch=self.operands)


class WORKERID(ZeroaryOperator):
    def __str__(self):
        return "%s" % self.__class__.__name__

    def __repr__(self):
        return "{op}()".format(op=self.opname())

    def evaluate(self, _tuple, scheme, state=None):
        return 0

    def typeof(self, scheme, state_scheme):
        return types.LONG_TYPE


class UnaryDoubleFunction(UnaryFunction):
    """A unary function that returns a double."""
    def typeof(self, scheme, state_scheme):
        input_type = self.input.typeof(scheme, state_scheme)
        check_is_numeric(input_type)
        return types.DOUBLE_TYPE


class UnaryTypePreservingFunction(UnaryFunction):
    def typeof(self, scheme, state_scheme):
        input_type = self.input.typeof(scheme, state_scheme)
        check_is_numeric(input_type)
        return input_type


class ABS(UnaryTypePreservingFunction):
    def evaluate(self, _tuple, scheme, state=None):
        return abs(self.input.evaluate(_tuple, scheme, state))


class CEIL(UnaryDoubleFunction):
    def evaluate(self, _tuple, scheme, state=None):
        return math.ceil(self.input.evaluate(_tuple, scheme, state))


class COS(UnaryDoubleFunction):
    def evaluate(self, _tuple, scheme, state=None):
        return math.cos(self.input.evaluate(_tuple, scheme, state))


class FLOOR(UnaryDoubleFunction):
    def evaluate(self, _tuple, scheme, state=None):
        return math.floor(self.input.evaluate(_tuple, scheme, state))


class LOG(UnaryDoubleFunction):
    def evaluate(self, _tuple, scheme, state=None):
        return math.log(self.input.evaluate(_tuple, scheme, state))


class SIN(UnaryDoubleFunction):
    def evaluate(self, _tuple, scheme, state=None):
        return math.sin(self.input.evaluate(_tuple, scheme, state))


class SQRT(UnaryDoubleFunction):
    def evaluate(self, _tuple, scheme, state=None):
        return math.sqrt(self.input.evaluate(_tuple, scheme, state))


class TAN(UnaryDoubleFunction):
    def evaluate(self, _tuple, scheme, state=None):
        return math.tan(self.input.evaluate(_tuple, scheme, state))


class MD5(UnaryFunction):
    def typeof(self, scheme, state_scheme):
        return types.LONG_TYPE

    def evaluate(self, _tuple, scheme, state=None):
        """Preserve 64 bits of the md5 hash function."""
        m = md5.new()
        m.update(str(self.input.evaluate(_tuple, scheme, state)))
        return int(m.hexdigest(), 16) >> 64


class POW(BinaryFunction):
    literals = ['POW']

    def evaluate(self, _tuple, scheme, state=None):
        return pow(self.left.evaluate(_tuple, scheme, state),
                   self.right.evaluate(_tuple, scheme, state))

    def typeof(self, scheme, state_scheme):
        lt = self.left.typeof(scheme, state_scheme)
        check_is_numeric(lt)
        rt = self.right.typeof(scheme, state_scheme)
        check_is_numeric(rt)

        return types.DOUBLE_TYPE


class CompareFunction(BinaryFunction):
    def typeof(self, scheme, state_scheme):
        lt = self.left.typeof(scheme, state_scheme)
        rt = self.right.typeof(scheme, state_scheme)
        if lt != rt:
            raise TypeSafetyViolation("Can't compare %s with %s" % (
                lt, rt))
        return lt


class LESSER(CompareFunction):
    literals = ['LESSER']

    def evaluate(self, _tuple, scheme, state=None):
        return min(self.left.evaluate(_tuple, scheme, state),
                   self.right.evaluate(_tuple, scheme, state))


class GREATER(CompareFunction):
    literals = ['GREATER']

    def evaluate(self, _tuple, scheme, state=None):
        return max(self.left.evaluate(_tuple, scheme, state),
                   self.right.evaluate(_tuple, scheme, state))


class SUBSTR(NaryFunction):
    literals = ["SUBSTR"]

    def evaluate(self, _tuple, scheme, state=None):
        inputStr = self.operands[0].evaluate(_tuple, scheme, state)
        beginIdx = self.operands[1].evaluate(_tuple, scheme, state)
        endIdx = self.operands[2].evaluate(_tuple, scheme, state)
        return inputStr[beginIdx:endIdx]

    def typeof(self, scheme, state_scheme):
        check_type(self.operands[0].typeof(scheme, state_scheme), types.STRING_TYPE)  # noqa
        check_type(self.operands[1].typeof(scheme, state_scheme), types.LONG_TYPE)  # noqa
        check_type(self.operands[2].typeof(scheme, state_scheme), types.LONG_TYPE)  # noqa

        return types.STRING_TYPE


class LEN(UnaryFunction):
    literals = ["LEN"]

    def evaluate(self, _tuple, scheme, state=None):
        return len(self.input.evaluate(_tuple, scheme, state))

    def typeof(self, scheme, state_scheme):
        input_type = self.input.typeof(scheme, state_scheme)
        if input_type != types.STRING_TYPE:
            raise TypeSafetyViolation("Must be a string for %s" % (
                self.__class__,))
        return types.LONG_TYPE



from raco.algebra import *
from raco.myrial.exceptions import MyrialCompileException

import bisect
import copy
import itertools
import logging
import networkx as nx

"""Control flow graph implementation.

Nodes are operations, edges are control flow.
"""

LOG = logging.getLogger(__name__)


def find_gt(a, x):
    """Find smallest value strictly greater than x"""
    i = bisect.bisect(a, x)
    if i != len(a):
        return a[i]
    return None


def find_lt(a, x):
    """Find largest value strictly less than x"""
    i = bisect.bisect(a, x)
    if i - 1 > 0:
        return a[i - 1]
    return None


def sliding_window(seq, n=2):
    """Returns a sliding window (of width n) over data from the iterable.

    http://stackoverflow.com/questions/6822725/rolling-or-sliding-window-iterator-in-python  # noqa
    """

    it = iter(seq)
    result = tuple(itertools.islice(it, n))
    if len(result) == n:
        yield result
    for elem in it:
        result = result[1:] + (elem,)
        yield result


class ControlFlowGraph(object):
    def __init__(self):
        self.graph = nx.DiGraph()
        self.sorted_vertices = []
        self._next_op_id = 0

    def __str__(self):
        g = self.graph
        node_strs = ['%s: uses=%s def=%s' % (str(n), repr(g.node[n]['uses']),
                                             repr(g.node[n]['def_var']))
                     for n in g]
        edge_strs = ['%s=>%s' % (str(s), str(d)) for s, d in g.edges()]
        return '; '.join(node_strs) + '\n' + '; '.join(edge_strs)

    @property
    def next_op_id(self):
        return self._next_op_id

    def add_op(self, op, _def, uses_set):
        """Add an operation to the CFG.

        :param op: A relational algebra operation (plan)
        :type op: raco.algebra.Operator
        :param _def: The variable defined by the operation or None
        :type _def: string
        :param uses_set: Set of variables read by the operation
        :type uses_set: Set of strings
        """

        op_id = self.next_op_id
        self._next_op_id += 1

        self.graph.add_node(op_id, op=op, def_var=_def, uses=uses_set)
        self.sorted_vertices.append(op_id)

        # Add a control flow edge from the previous statement; this assumes we
        # don't do jumps or any other non-linear control flow.
        if op_id > 0:
            self.graph.add_edge(op_id - 1, op_id)

    def add_edge(self, source, dest):
        """An an explicit edge to the control flow graph."""
        self.graph.add_edge(source, dest)

    def compute_liveness(self):
        """Run liveness analysis over the control flow graph.

        http://www.cs.colostate.edu/~mstrout/CS553/slides/lecture03.pdf

        :returns: A tuple containing live_in, live_out dictionaries.  The keys
        are variable names (strings) and the values are string sets.
        """

        # All variables that are accessed are live-in at a node
        live_in = {i: copy.copy(self.graph.node[i]['uses'])
                   for i in self.graph}
        live_out = {i: set() for i in self.graph}

        while True:
            # Create a copy of live_in, live_out that we will compare against
            # to detect loop termination. This is shallower than a full deep
            # copy, but we need to copy the value sets in this dictionary as
            # they are mutated in the loop below. Make the sets frozen to
            # ensure they are really immutable.
            live_in_prev = {i: frozenset(s) for i, s in live_in.items()}
            live_out_prev = {i: frozenset(s) for i, s in live_out.items()}

            for i in self.graph:
                # live out variables that are not defined are live-in
                def_var = self.graph.node[i]['def_var']
                def_set = set()
                if def_var is not None:
                    def_set.add(def_var)
                live_in[i].update(live_out_prev[i] - def_set)

                # variables that are live-in at a successor are live-out
                for successor in self.graph.successors(i):
                    live_out[i].update(live_in_prev[successor])

            if live_in == live_in_prev and live_out == live_out_prev:
                return live_in, live_out

    def __delete_node(self, node):
        """Remove a node from the control flow graph.

        Add an edge to the graph to "skip over" the target node.
        """
        assert node in self.graph

        # find a successor of the given node
        predecessors = self.graph.predecessors(node)
        if predecessors:
            successor = find_gt(self.sorted_vertices, node)
            if successor is not None:
                for p in predecessors:
                    self.graph.add_edge(p, successor)

        self.graph.remove_node(node)
        self.sorted_vertices.remove(node)

        assert node not in self.graph

    def __inline_node(self, dest_node, target_node):
        """Inline the target node into the destination node."""

        assert target_node in self.graph
        assert dest_node in self.graph

        target_op = self.graph.node[target_node]['op']
        assert isinstance(target_op, StoreTemp)
        target_inner_op = target_op.input

        dest_op = self.graph.node[dest_node]['op']

        var = self.graph.node[target_node]['def_var']
        assert var is not None

        new_op = inline_operator(dest_op, var, target_inner_op)
        self.graph.node[dest_node]['op'] = new_op

        # The merged node uses the union of the previous nodes input variables
        uses_set = self.graph.node[dest_node]['uses']
        uses_set.remove(var)
        uses_set.update(self.graph.node[target_node]['uses'])

        self.__delete_node(target_node)

    def apply_chaining(self):
        """Merge adjacent statements by chaining together plans.

        It is often desirable to chain plans instead of materializing temporary
        tables. Consider this simple example:

        X = SCAN(foo);   -- Materializing a temporary table for X would be dumb
        Y = DISTINCT(X); -- Instead, we can inline the scan into this
                         -- expression

        The merge procedure operates on the control flow graph.  We inline node
        A into node B whenever the following conditions are all true:

        - A directly precedes B; we don't consider out-of-order executions
        - A defines a variable (i.e., it assigns a temporary; not a STORE)
        - B references the variable defined by A -- def(A) in uses(B)
        - The variable defined by A is not used again
              -- def(A) not in live_out(B)
        - A and B are in the same do/while loop.

        The merge procedure is applied recursively on the CFG until convergence
        is reached.
        """

        _continue = True
        while _continue:
            live_in, live_out = self.compute_liveness()
            _continue = False

            # TODO XXX O(N^2) algorithm

            # Walk through the program backwards, and try inlining line A into
            # line B according to the above logic.
            #
            # In most cases, A is line i and B is line i+1. However, when we
            # successfully inline A into B, we want to consider inlining the
            # next A' into B, not into A.  In these cases we save the current B
            # in the inlined_into variable and then reuse it as the inline
            # candidate the next round.
            inlined_into = None
            for nodeB, nodeA in sliding_window(reversed(self.sorted_vertices)):
                if inlined_into is not None:
                    nodeB = inlined_into
                    inlined_into = None

                if self.graph.in_degree(nodeB) == 2:
                    continue  # start of do/while loop

                def_var = self.graph.node[nodeA]['def_var']
                if not def_var:
                    continue

                uses = self.graph.node[nodeB]['uses']
                if def_var not in uses:
                    continue

                if def_var in live_out[nodeB]:
                    continue

                self.__inline_node(nodeB, nodeA)
                _continue = True
                inlined_into = nodeB

    def dead_code_elimination(self):
        """Dead code elimination.

        Specifically: delete CFG nodes that define a variable that is not in
        the live_out set. Recurse until convergence.
        """

        _continue = True
        while _continue:
            _continue = False
            live_in, live_out = self.compute_liveness()

            for node in self.graph:
                out_set = live_out[node]
                def_var = self.graph.node[node]['def_var']

                # Only delete nodes that 1) Define a variable (and therefore
                # aren't STORE, etc.); 2) Are not required downstream.
                if def_var and def_var not in out_set:
                    self.__delete_node(node)
                    _continue = True
                    break

    def dead_loop_elimination(self):
        """Delete entire do/while loops whose results are not consumed.

        See get_logical_plan for logic."""

        if len(self.sorted_vertices) == 0:
            return

        # A stack that contains the defined variables within each loop.
        def_set_stack = []

        def current_def_set():
            if len(def_set_stack) > 0:
                return def_set_stack[-1]
            else:
                return None

        current_loop_first_index = -1
        loops_to_delete = []  # tuples of the form [begin_index, end_index]

        live_in, live_out = self.compute_liveness()
        last_op = self.sorted_vertices[-1]

        for i in self.sorted_vertices:
            if self.graph.in_degree(i) == 2:
                # start new do/while loop
                current_loop_first_index = i
                def_set_stack.append(set())
                # Add anything defined by the current statement to the def_set
                def_var = self.graph.node[i]['def_var']
                if def_var:
                    def_set_stack[-1].add(def_var)
            elif (current_def_set() is not None and
                    (self.graph.out_degree(i) == 2 or i == last_op)):
                # end of do/while loop: check whether anything this loop
                # defines is live_in after the loop.
                def_set = def_set_stack.pop()
                next_op = find_gt(self.sorted_vertices, i)

                loop_range = (current_loop_first_index, i)

                if next_op is None:
                    # no next node?  Loop is obviously dead
                    loops_to_delete.append(loop_range)
                elif len(def_set.intersection(live_in[next_op])) == 0:
                    loops_to_delete.append(loop_range)
            elif current_def_set() is not None:
                # Add anything defined by the current statement to the def_set
                def_var = self.graph.node[i]['def_var']
                if def_var:
                    def_set_stack[-1].add(def_var)

        if not loops_to_delete:
            return

        # Delete the operations corresponding to dead loops
        for begin, end in loops_to_delete:
            for ix in range(begin, end + 1):
                self.graph.remove_node(ix)
                self.sorted_vertices.remove(ix)

            # Add a control flow edge that "skips over" the deleted loop
            if begin > 0 and end < last_op:
                prev = find_lt(self.sorted_vertices, begin)
                _next = find_gt(self.sorted_vertices, end)

                assert prev in self.graph
                assert _next in self.graph
                assert prev < begin
                assert _next > end
                self.graph.add_edge(prev, _next)

        self.dead_loop_elimination()

    def get_logical_plan(self, dead_code_elimination=True,
                         apply_chaining=True):
        """Extract a logical plan from the control flow graph.

        The logic here is simplistic:
        * Any node with in-degree == 2 is the top of a new do/while loop
        * Any node with out-degree == 2 is a while condition
        * Any other node is an ordinary operation.

        :returns: An instance of raco.algebra.Operator
        """

        if dead_code_elimination:
            self.dead_loop_elimination()
            self.dead_code_elimination()

        if apply_chaining:
            self.apply_chaining()

        if not self.graph:
            raise MyrialCompileException("Optimized program is empty")

        op_stack = [Sequence()]

        def current_block():
            return op_stack[-1]

        for i in sorted(self.graph):
            op = self.graph.node[i]['op']

            if self.graph.out_degree(i) == 2:
                LOG.info("Terminating while loop (%d): %s", i, op)
                # Terminate current do/while loop
                assert isinstance(current_block(), DoWhile)
                current_block().add(op)

                do_while_op = op_stack.pop()
                current_block().add(do_while_op)
                continue

            if self.graph.in_degree(i) == 2:
                LOG.info("Introducing while loop (%d)", i)
                # Introduce new do/while loop
                op_stack.append(DoWhile())

            LOG.info("Adding operation to sequence (%d) %s", i, op)
            current_block().add(op)

        assert len(op_stack) == 1
        return current_block()



"""Basic test of the command-line interface to Myrial."""

import subprocess
import unittest


class CliTest(unittest.TestCase):

    def test_cli(self):
        out = subprocess.check_output(['python', 'scripts/myrial',
                                       'examples/reachable.myl'])
        self.assertIn('DO', out)
        self.assertIn('WHILE', out)

    def test_cli_standalone_execute(self):
        out = subprocess.check_output(['python', 'scripts/myrial', '-f',
                                       'examples/standalone.myl'])
        self.assertIn('Dan Suciu,engineering', out)

    def test_cli_standalone_json(self):
        out = subprocess.check_output(['python', 'scripts/myrial', '-j',
                                       'examples/cast.myl'])
        self.assertIn('rawQuery', out)

    def test_cli_standalone_logical(self):
        out = subprocess.check_output(['python', 'scripts/myrial', '-l',
                                       'examples/standalone.myl'])
        self.assertIn("CrossProduct[FileScan", out)

    def test_cli_standalone_repr(self):
        out = subprocess.check_output(['python', 'scripts/myrial', '-r',
                                       'examples/standalone.myl'])
        self.assertIn("FileScan('./examples/dept.csv'", out)

    def test_cli_reserved_column_name(self):
        proc = subprocess.Popen(
            ['python', 'scripts/myrial', 'examples/bad_column_name.myl'],
            stdout=subprocess.PIPE)
        out = proc.communicate()[0]
        self.assertIn('The token "SafeDiv" on line 2 is reserved', out)



"""Functionality for handling queries with multiple relational arguments."""

from raco import algebra
from raco import expression
from raco.expression.statevar import *

from raco.myrial.exceptions import ColumnIndexOutOfBounds


def rewrite_statemods(statemods, from_args, base_offsets):
    """Convert Unbox expressions contained inside statemod variables.

    :param statemods: A list of StateVar instances
    :param from_args: A map from relation name to Operator
    :param base_offsets: A map from relation name to initial column offsets
    :return: An updated list of StateVar instances
    """
    assert all(isinstance(sm, StateVar) for sm in statemods)
    return [StateVar(name, init, rewrite_refs(update, from_args, base_offsets))
            for name, init, update in statemods]


def rewrite_refs(sexpr, from_args, base_offsets):
    """Convert all Unbox expressions into raw indexes."""

    def rewrite_node(sexpr):
        # Push unboxing into the state variables of distributed aggregates
        if isinstance(sexpr, expression.AggregateExpression):
            if sexpr.is_decomposable():
                ds = sexpr.get_decomposable_state()
                lsms = rewrite_statemods(ds.get_local_statemods(), from_args, base_offsets)  # noqa
                rsms = rewrite_statemods(ds.get_remote_statemods(), from_args, base_offsets)  # noqa

                if lsms or rsms:
                    sexpr.set_decomposable_state(
                        expression.DecomposableAggregateState(
                            ds.get_local_emitters(), lsms,
                            ds.get_remote_emitters(), rsms,
                            ds.get_finalizer()))
                return sexpr

        if not isinstance(sexpr, expression.Unbox):
            return sexpr
        else:
            op = from_args[sexpr.relational_expression]
            scheme = op.scheme()

            debug_info = None
            if not sexpr.field:
                offset = 0
            elif isinstance(sexpr.field, int):
                if sexpr.field >= len(scheme):
                    raise ColumnIndexOutOfBounds(str(sexpr))
                offset = sexpr.field
            else:
                assert isinstance(sexpr.field, basestring)
                offset = scheme.getPosition(sexpr.field)
                debug_info = sexpr.field

            offset += base_offsets[sexpr.relational_expression]
            return expression.UnnamedAttributeRef(offset, debug_info)

    def recursive_eval(sexpr):
        """Rewrite a node and all its descendents"""
        newexpr = rewrite_node(sexpr)
        newexpr.apply(recursive_eval)
        return newexpr

    return recursive_eval(sexpr)


def __calculate_offsets(from_args):
    """Calculate the first column of each relation in the rollup schema."""
    index = 0
    offsets = {}
    for _id in from_args.iterkeys():
        offsets[_id] = index
        index += len(from_args[_id].scheme())

    return offsets


def merge(from_args):
    """Merge a sequence of operations into a cross-product tree.

    from_args: A dictionary mapping a unique string id to a
    raco.algebra.Operation instance.

    Returns: a single raco.algebra.Operation instance and an opaque
    data structure suitable for passing to the rewrite_refs function.
    """

    assert len(from_args) > 0

    def cross(x, y):
        return algebra.CrossProduct(x, y)

    from_ops = from_args.values()
    op = reduce(cross, from_ops)

    return (op, __calculate_offsets(from_args))



from abc import abstractmethod, ABCMeta
from raco.algebra import DEFAULT_CARDINALITY
from raco.relation_key import RelationKey
from raco.scheme import Scheme
from ast import literal_eval


class Relation(object):
    def __init__(self, name, sch):
        self.name = name
        self._scheme = sch

    def __eq__(self, other):
        return self.name == other.name
        # and self.scheme == other.scheme

    def scheme(self):
        return self._scheme


class FileRelation(Relation):
    pass


class ASCIIFile(FileRelation):
    pass


class Catalog(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def get_num_servers(self):
        """ Return number of servers in myria deployment """

    @abstractmethod
    def get_scheme(self, rel_key):
        """ Return scheme of tuples of rel_key """

    @abstractmethod
    def num_tuples(self, rel_key):
        """ Return number of tuples of rel_key """


# Some useful Catalog implementations

class FakeCatalog(Catalog):
    """ fake catalog, should only be used in test """
    def __init__(self, num_servers, child_sizes=None):
        self.num_servers = num_servers
        # default sizes
        self.cached = {}
        # overwrite default sizes if necessary
        if child_sizes:
            for child, size in child_sizes.items():
                self.cached[RelationKey(child)] = size

    def get_num_servers(self):
        return self.num_servers

    def num_tuples(self, rel_key):
        if rel_key in self.cached:
            return self.cached[rel_key]
        return DEFAULT_CARDINALITY

    def get_scheme(self, rel_key):
        raise NotImplementedError()


class FromFileCatalog(Catalog):
    """ Catalog that is created from a python file.
    Format of file is a dictionary of schemas.

    {'relation1' : [('a', 'LONG_TYPE'), ('b', 'STRING_TYPE')],
     'relation2' : [('y', 'STRING_TYPE'), ('z', 'DATETIME_TYPE')]}

     Or there can be an optional cardinality for any relation
    {'relation1' : ([('a', 'LONG_TYPE'), ('b', 'STRING_TYPE')], 10),
     'relation2' : [('y', 'STRING_TYPE'), ('z', 'DATETIME_TYPE')]}

     see raco.types for allowed types
    """

    def __init__(self, cat):
        self.catalog = {}

        def parse(v):
            if isinstance(v, tuple):
                return v
            elif isinstance(v, list):
                return v, DEFAULT_CARDINALITY
            else:
                assert False, """Unexpected catalog file format. \
                See raco.catalog.FromFileCatalog"""

        self.catalog = dict([(k, parse(v)) for k, v in cat.iteritems()])

    def get_scheme(self, rel_key):
        return Scheme(self.__get_catalog_entry__(rel_key)[0])

    @classmethod
    def load_from_file(cls, path):
        with open(path) as fh:
            return cls(literal_eval(fh.read()))

    def get_num_servers(self):
        return 1

    def __get_catalog_entry__(self, rel_key):
        e = self.catalog.get(str(rel_key))
        if e is None:
            raise Exception(
                "relation {r} not found in catalog".format(r=rel_key))

        return e

    def num_tuples(self, rel_key):
        return self.__get_catalog_entry__(rel_key)[1]




import collections
import itertools
import csv

from raco.dbconn import DBConnection
from raco import relation_key, types
from raco.algebra import StoreTemp, DEFAULT_CARDINALITY
from raco.catalog import Catalog
from raco.expression import AND, EQ, BuiltinAggregateExpression

debug = False


class State(object):
    def __init__(self, op_scheme, state_scheme, init_exprs):
        self.scheme = state_scheme
        self.op_scheme = op_scheme
        self.values = [x.evaluate(None, op_scheme, None)
                       for (_, x) in init_exprs]

    def update(self, tpl, update_exprs):
        new_vals = [expr.evaluate(tpl, self.op_scheme, self)
                    for (_, expr) in update_exprs]
        self.values = new_vals

    def __str__(self):
        return 'State(%s)' % self.values


class FakeDatabase(Catalog):
    """An in-memory implementation of relational algebra operators"""

    def __init__(self):
        # Persistent tables, identified by RelationKey
        self.tables = DBConnection()

        # Temporary tables, identified by string name
        self.temp_tables = DBConnection()

    def get_num_servers(self):
        return 1

    def num_tuples(self, rel_key):
        try:
            return self.tables.num_tuples(rel_key)
        except KeyError:
            return DEFAULT_CARDINALITY

    def evaluate(self, op):
        """Evaluate a relational algebra operation.

        For "query-type" operators, return a tuple iterator.
        For store queries, the return value is None.
        """
        method = getattr(self, op.opname().lower())
        return method(op)

    def evaluate_to_bag(self, op):
        """Return a bag (collections.Counter instance) for the operation"""
        return collections.Counter(self.evaluate(op))

    def ingest(self, rel_key, contents, scheme):
        """Directly load raw data into the database"""
        if isinstance(rel_key, str):
            rel_key = relation_key.RelationKey.from_string(rel_key)
        assert isinstance(rel_key, relation_key.RelationKey)
        self.tables.add_table(rel_key, scheme, contents.elements())

    def get_scheme(self, rel_key):
        if isinstance(rel_key, str):
            rel_key = relation_key.RelationKey.from_string(rel_key)

        assert isinstance(rel_key, relation_key.RelationKey)

        return self.tables.get_scheme(rel_key)

    def get_table(self, rel_key):
        """Retrieve the contents of table.

        :param rel_key: The key of the relation
        :type rel_key: relation_key.RelationKey
        :returns: A collections.Counter instance containing tuples.
        """
        if isinstance(rel_key, str):
            rel_key = relation_key.RelationKey.from_string(rel_key)
        assert isinstance(rel_key, relation_key.RelationKey)
        return self.tables.get_table(rel_key)

    def get_temp_table(self, key):
        return self.temp_tables.get_table(key)

    def delete_temp_table(self, key):
        self.temp_tables.delete_table(key)

    def dump_all(self):
        for key, val in self.tables.iteritems():
            bag = val[0]
            print '%s: (%s)' % (key, bag)

        for key, bag in self.temp_tables.iteritems():
            print '__%s: (%s)' % (key, bag)

    def scan(self, op):
        assert isinstance(op.relation_key, relation_key.RelationKey)
        return self.tables.get_table(op.relation_key).elements()

    def filescan(self, op):
        type_list = op.scheme().get_types()

        with open(op.path, 'r') as fh:
            sample = fh.read(1024)
            dialect = csv.Sniffer().sniff(sample)
            fh.seek(0)
            reader = csv.reader(fh, dialect)
            for row in reader:
                pairs = zip(row, type_list)
                cols = [types.parse_string(s, t) for s, t in pairs]
                yield tuple(cols)

    def select(self, op):
        child_it = self.evaluate(op.input)

        def filter_func(_tuple):
            # Note: this implicitly uses python truthiness rules for
            # interpreting non-boolean expressions.
            # TODO: Is this the the right semantics here?
            return op.condition.evaluate(_tuple, op.scheme())

        return itertools.ifilter(filter_func, child_it)

    def apply(self, op):
        child_it = self.evaluate(op.input)
        scheme = op.input.scheme()

        def make_tuple(input_tuple):
            ls = [colexpr.evaluate(input_tuple, scheme)
                  for (_, colexpr) in op.emitters]
            return tuple(ls)
        return (make_tuple(t) for t in child_it)

    def statefulapply(self, op):
        child_it = self.evaluate(op.input)
        scheme = op.input.scheme()

        state = State(scheme, op.state_scheme, op.inits)

        def make_tuple(input_tuple, state):
            # Update state variables
            state.update(input_tuple, op.updaters)

            # Extract a result for each emit expression
            return tuple([colexpr.evaluate(input_tuple, scheme, state)
                          for (_, colexpr) in op.emitters])

        return (make_tuple(t, state) for t in child_it)

    def join(self, op):
        # Compute the cross product of the children and flatten
        left_it = self.evaluate(op.left)
        right_it = self.evaluate(op.right)
        p1 = itertools.product(left_it, right_it)
        p2 = (x + y for (x, y) in p1)

        # Return tuples that match on the join conditions
        return (tpl for tpl in p2 if op.condition.evaluate(tpl, op.scheme()))

    def projectingjoin(self, op):
        # standard join, projecting the output columns
        return (tuple(t[x.position] for x in op.output_columns)
                for t in self.join(op))

    def naryjoin(self, op):
        def eval_conditions(conditions, tpl):
            """Turns the weird NaryJoin condition set into a proper
            expression, then evaluates it."""
            cond = reduce(lambda a, b: AND(a, b),
                          map(lambda (a, b): EQ(a, b), conditions))
            return cond.evaluate(tpl, op.scheme())

        # Elements of prod are tuples of tuples like ((1, 2), (3, 4))
        prod = itertools.product(*(self.evaluate(child)
                                   for child in op.children()))
        # Elements of tuples have been flattened like (1, 2, 3, 4)
        tuples = (sum(x, ()) for x in prod)
        return (tpl for tpl in tuples if eval_conditions(op.conditions, tpl))

    def crossproduct(self, op):
        left_it = self.evaluate(op.left)
        right_it = self.evaluate(op.right)
        p1 = itertools.product(left_it, right_it)
        return (x + y for (x, y) in p1)

    def distinct(self, op):
        it = self.evaluate(op.input)
        s = set(it)
        return iter(s)

    def project(self, op):
        if not op.columnlist:
            return self.distinct(op)

        return set(tuple(t[x.position] for x in op.columnlist)
                   for t in self.evaluate(op.input))

    def limit(self, op):
        it = self.evaluate(op.input)
        return itertools.islice(it, op.count)

    @staticmethod
    def singletonrelation(op):
        return iter([()])

    @staticmethod
    def emptyrelation(op):
        return iter([])

    def unionall(self, op):
        left_it = self.evaluate(op.left)
        right_it = self.evaluate(op.right)
        return itertools.chain(left_it, right_it)

    def union(self, op):
        return set(x for x in self.unionall(op))

    def difference(self, op):
        its = [self.evaluate(op.left), self.evaluate(op.right)]
        sets = [set(it) for it in its]
        return sets[0].difference(sets[1])

    def intersection(self, op):
        its = [self.evaluate(op.left), self.evaluate(op.right)]
        sets = [set(it) for it in its]
        return sets[0].intersection(sets[1])

    def groupby(self, op):
        child_it = self.evaluate(op.input)
        input_scheme = op.input.scheme()

        def process_grouping_columns(_tuple):
            ls = [sexpr.evaluate(_tuple, input_scheme) for
                  sexpr in op.grouping_list]
            return tuple(ls)

        # Calculate groups of matching input tuples.
        # If there are no grouping terms, then all tuples are added
        # to a single bin.
        results = collections.defaultdict(list)

        if len(op.grouping_list) == 0:
            results[()] = list(child_it)
        else:
            for input_tuple in child_it:
                grouped_tuple = process_grouping_columns(input_tuple)
                results[grouped_tuple].append(input_tuple)

        # resolve aggregate functions
        for key, tuples in results.iteritems():
            state = State(input_scheme, op.state_scheme, op.inits)
            for tpl in tuples:
                state.update(tpl, op.updaters)

            # For now, built-in aggregates are handled differently than UDA
            # aggregates.  TODO: clean this up!

            agg_fields = []
            for expr in op.aggregate_list:
                if isinstance(expr, BuiltinAggregateExpression):
                    # Old-style aggregate: pass all tuples to the eval func
                    agg_fields.append(
                        expr.evaluate_aggregate(tuples, input_scheme))
                else:
                    # UDA-style aggregate: evaluate a normal expression that
                    # can reference only the state tuple
                    agg_fields.append(expr.evaluate(None, None, state))
            yield(key + tuple(agg_fields))

    def sequence(self, op):
        for child_op in op.children():
            self.evaluate(child_op)
        return None

    def parallel(self, op):
        for child_op in op.children():
            self.evaluate(child_op)
        return None

    def dowhile(self, op):
        i = 0

        children = op.children()
        body_ops = children[:-1]
        term_op = children[-1]
        if isinstance(term_op, StoreTemp):
            term_op = term_op.input

        if debug:
            print '---------- Values at top of do/while -----'
            self.dump_all()

        while True:
            for op in body_ops:
                self.evaluate(op)
            result_iterator = self.evaluate(term_op)

            if debug:
                i += 1
                print '-------- Iteration %d ------------' % i
                self.dump_all()

            try:
                tpl = result_iterator.next()

                if debug:
                    print 'Term: %s' % str(tpl)

                # XXX should we use python truthiness here?
                if not tpl[0]:
                    break
            except StopIteration:
                break
            except IndexError:
                break

    def store(self, op):
        assert isinstance(op.relation_key, relation_key.RelationKey)

        scheme = op.input.scheme()
        self.tables.add_table(op.relation_key, scheme, self.evaluate(op.input))
        return None

    def dump(self, op):
        for tpl in self.evaluate(op.input):
            print ','.join(tpl)
        return None

    def storetemp(self, op):
        scheme = op.input.scheme()
        self.temp_tables.add_table(op.name, scheme, self.evaluate(op.input))

    def appendtemp(self, op):
        self.temp_tables.append_table(op.name, self.evaluate(op.input))

    def scantemp(self, op):
        return self.temp_tables.get_table(op.name).elements()

    def myriascan(self, op):
        return self.scan(op)

    def myriascantemp(self, op):
        return self.scantemp(op)

    def myriasymmetrichashjoin(self, op):
        return self.projectingjoin(op)

    def myrialeapfrogjoin(self, op):
        # standard naryjoin, projecting the output columns
        return (tuple(t[x.position] for x in op.output_columns)
                for t in self.naryjoin(op))

    def myriainmemoryorderby(self, op):
        return self.evaluate(op.input)

    def myriahypershuffleconsumer(self, op):
        return self.evaluate(op.input)

    def myriahypershuffleproducer(self, op):
        return self.evaluate(op.input)

    def myriastore(self, op):
        return self.store(op)

    def myriastoretemp(self, op):
        return self.storetemp(op)

    def myriaappendtemp(self, op):
        return self.appendtemp(op)

    def myriaapply(self, op):
        return self.apply(op)

    def myriastatefulapply(self, op):
        return self.statefulapply(op)

    def myriadupelim(self, op):
        return self.distinct(op)

    def myriaselect(self, op):
        return self.select(op)

    def myriacrossproduct(self, op):
        return self.crossproduct(op)

    def myriagroupby(self, op):
        return self.groupby(op)

    def myriashuffleconsumer(self, op):
        return self.evaluate(op.input)

    def myriashuffleproducer(self, op):
        return self.evaluate(op.input)

    def myriacollectconsumer(self, op):
        return self.evaluate(op.input)

    def myriacollectproducer(self, op):
        return self.evaluate(op.input)

    def myriabroadcastconsumer(self, op):
        return self.evaluate(op.input)

    def myriabroadcastproducer(self, op):
        return self.evaluate(op.input)

    def myriasingleton(self, op):
        return self.singletonrelation(op)

    def myriaemptyrelation(self, op):
        return self.emptyrelation(op)

    def myriaunionall(self, op):
        return self.unionall(op)

    def myriadifference(self, op):
        return self.difference(op)

    def myriaqueryscan(self, op):
        return self.tables.get_sql_output(op.sql).elements()



from nose.plugins.skip import SkipTest
import unittest
import algebra
from raco import RACompiler
from raco.catalog import FakeCatalog
from raco.language import myrialang


class testNaryJoin(unittest.TestCase):
    """ Unit tests for Nary Join related rules"""
    @staticmethod
    def get_phys_plan_root(query, num_server, child_size=None):
        dlog = RACompiler()
        dlog.fromDatalog(query)
        dlog.optimize(myrialang.MyriaHyperCubeAlgebra(
            FakeCatalog(num_server, child_size)))
        ret = dlog.physicalplan
        assert isinstance(ret, algebra.Parallel)
        assert len(ret.children()) == 1
        ret = ret.children()[0]
        assert isinstance(ret, algebra.Store)
        return ret.input

    @SkipTest
    def test_merge_to_nary_join(self):
        """ Test the rule merging binary join to nary join.
        """
        def in_cond_idx(position, conditions):
            for i, cond in enumerate(conditions):
                if position in cond:
                    return i
            raise Exception("Cannot find attribute in join conditions")

        # 1. triangular join
        triangle_join = testNaryJoin.get_phys_plan_root(
            "A(x,y,z):-R(x,y),S(y,z),T(z,x)", 64)
        # test root operator type
        self.assertIsInstance(triangle_join, algebra.NaryJoin)
        # test arity of join conditions
        self.assertEqual(len(triangle_join.conditions), 3)
        # test join conditions
        conds = []
        for cond in triangle_join.conditions:
            conds.append([attribute.position for attribute in cond])
        self.assertEqual(in_cond_idx(0, conds), in_cond_idx(5, conds))
        self.assertEqual(in_cond_idx(1, conds), in_cond_idx(2, conds))
        self.assertEqual(in_cond_idx(3, conds), in_cond_idx(4, conds))

        # 2. star join
        star_join = testNaryJoin.get_phys_plan_root(
            "A(x,y,z,p):-R(x,y),S(x,z),T(x,p)", 64)
        # test root operator type
        self.assertIsInstance(star_join, algebra.NaryJoin)
        # test arity of join conditions
        self.assertEqual(len(star_join.conditions), 1)
        # test join conditions
        conds = []
        for cond in star_join.conditions:
            conds.append([attribute.position for attribute in cond])
        self.assertEqual(in_cond_idx(0, conds), in_cond_idx(2, conds))
        self.assertEqual(in_cond_idx(2, conds), in_cond_idx(4, conds))

    @SkipTest
    def test_hashed_column_mapping(self):
        """Test whether hashed columns are mapped to correct HC dimensions."""
        def get_hc_dim(expr, column):
            """Return the mapped dimension in hyper cube."""
            return expr.mapped_hc_dimensions[expr.hashed_columns.index(column)]

        def get_shuffle_producers(expr):
            """ Go two steps deeper in the tree, find shuffle producers.
            """
            ret = [child.input.input for child in expr.children()]
            return tuple(ret)

        # 1. triangular join
        triangular_join = testNaryJoin.get_phys_plan_root(
            "A(x,y,z):-R(x,y),S(y,z),T(z,x)", 64)
        shuffle_r, shuffle_s, shuffle_t = get_shuffle_producers(
            triangular_join)
        # x in R and x in T are shuffled to the same dimension
        self.assertEqual(get_hc_dim(shuffle_r, 0), get_hc_dim(shuffle_t, 1))
        # y in R and y in S are shuffled to the same dimension
        self.assertEqual(get_hc_dim(shuffle_r, 1), get_hc_dim(shuffle_s, 0))
        # z in S and z in T are shuffled to the same dimension
        self.assertEqual(get_hc_dim(shuffle_s, 1), get_hc_dim(shuffle_t, 0))

        # 2. star join
        star_join = testNaryJoin.get_phys_plan_root(
            "A(x,y,z,p):-R(x,y),S(x,z),T(x,p)", 64)
        shuffle_r, shuffle_s, shuffle_t = get_shuffle_producers(star_join)
        # x in R and x in S are shuffled to the same dimension
        self.assertEqual(get_hc_dim(shuffle_r, 0), get_hc_dim(shuffle_s, 0))
        # x in S and x in T are shuffled to the same dimension
        self.assertEqual(get_hc_dim(shuffle_s, 0), get_hc_dim(shuffle_t, 0))

    @SkipTest
    def test_cell_partition(self):
        def get_cell_partition(expr, dim_sizes, child_idx):
            children = expr.children()
            children = [c.input.input for c in children]
            child_schemes = [c.scheme() for c in children]
            conditions = myrialang.convert_nary_conditions(
                expr.conditions, child_schemes)
            return myrialang.HCShuffleBeforeNaryJoin.get_cell_partition(
                dim_sizes, conditions, child_schemes,
                child_idx, children[child_idx].hashed_columns)

        # 1. triangular join
        expr = testNaryJoin.get_phys_plan_root(
            "A(x,y,z):-R(x,y),S(y,z),T(z,x)", 64, {"R": 1, "S": 100, "T": 20})
        dim_sizes = [1, 2, 2]
        # test cell partition of scan r
        self.assertEqual(
            get_cell_partition(expr, dim_sizes, 0), [[0, 1], [2, 3]])
        # test cell partition of scan s
        self.assertEqual(
            get_cell_partition(expr, dim_sizes, 1), [[0], [1], [2], [3]])
        # test cell partition of scan t
        self.assertEqual(
            get_cell_partition(expr, dim_sizes, 2), [[0, 2], [1, 3]])

        # 2. chain join
        expr = testNaryJoin.get_phys_plan_root(
            "A(x,y,z,p):-R(x,y),S(y,z),T(z,p)", 64)
        dim_sizes = [2, 2]
        self.assertEqual(
            get_cell_partition(expr, dim_sizes, 0), [[0, 1], [2, 3]])
        self.assertEqual(
            get_cell_partition(expr, dim_sizes, 1), [[0], [1], [2], [3]])
        self.assertEqual(
            get_cell_partition(expr, dim_sizes, 2), [[0, 2], [1, 3]])

    @SkipTest
    def test_dim_size(self):
        def get_dim_size(expr):
            producer = expr.children()[0].input.input
            return producer.hyper_cube_dimensions

        def get_work_load(expr, dim_sizes):
            children = expr.children()
            children = [c.input for c in children]
            child_schemes = [c.scheme() for c in children]
            conditions = myrialang.convert_nary_conditions(
                expr.conditions, child_schemes)
            HSClass = myrialang.HCShuffleBeforeNaryJoin
            r_index = HSClass.reversed_index(child_schemes, conditions)
            child_sizes = [len(cs) for cs in child_schemes]
            return HSClass.workload(dim_sizes, child_sizes, r_index)

        # test triangle join with equal input size
        triangular_join = testNaryJoin.get_phys_plan_root(
            "A(x,y,z):-R(x,y),S(y,z),T(z,x)", 64)
        self.assertEqual(get_dim_size(triangular_join), (4, 4, 4))

        # test rectange join with equal input size
        rect_join = testNaryJoin.get_phys_plan_root(
            "A(x,y,z):-R(x,y),S(y,z),T(z,p), N(p,x)", 256)
        # note: there is more than one optimal [4,4,4,4] or [1,16,1,16] etc.
        self.assertEqual(get_work_load(rect_join, [4, 4, 4, 4]),
                         get_work_load(rect_join, get_dim_size(rect_join)))



"""

classes for representing and manipulating Datalog programs.

In particular, they can be compiled to (iterative) relational algebra
expressions.
"""
import networkx as nx
from raco import expression
import raco.algebra as algebra
from raco.scheme import Scheme
import raco.catalog
import raco.myrial.groupby
from raco.relation_key import RelationKey
import raco.types

import logging
LOG = logging.getLogger(__name__)


def make_attr(i, r, relation_alias):
    if isinstance(r, Var):
        name = r.var
    elif isinstance(r, expression.Literal):
        name = "pos%s" % i
    attrtype = r.typeof(None, None)
    return (name, attrtype)


class Program(object):
    def __init__(self, rules):
        self.rules = rules
        self.compiledidbs = {}

    def isIDB(self, term):
        """Is this term also an IDB?"""
        return self.IDB(term) != []

    def IDB(self, term):
        """Return a list of rules that define an IDB corresponding to the given
        term: relation names are the same, and the number of columns are the
        same."""
        matches = []
        for r in self.rules:
            if r.IDBof(term):
                matches.append(r)

        return matches

    def compiling(self, idbterm):
        """Return true if any of the rules that define the given idb are being
        compiled"""
        for r in self.rules:
            if r.head.samerelation(idbterm):
                if r.compiling:
                    return True
        return False

    def intermediateRule(self, rule):
        """Return True if the head appears in the body of any other rule."""
        for other in self.rules:
            if other.refersTo(rule.head) and other.head != rule.head:
                return True
        return False

    def toRA(self):
        """Return a set of relational algebra expressions implementing this
        program."""
        self.idbs = {}
        for rule in self.rules:
            block = self.idbs.setdefault(rule.head.name, [])
            block.append(rule)

        return algebra.Parallel([algebra.Store(RelationKey(idb),
                                               self.compileIDB(idb))
                                 for (idb, rules) in self.idbs.items()
                                 if any([not self.intermediateRule(r)
                                         for r in rules])])

    def compileIDB(self, idb):
        """Compile an idb by name.  Uses the self.idbs data structure created
        in self.toRA"""

        if idb in self.compiledidbs:
            return self.compiledidbs[idb]
        else:
            rules = self.idbs[idb]
            plans = [r.toRA(self) for r in rules]
            ra = reduce(algebra.Union, plans)
            self.compiledidbs[self] = ra
            return ra

    def __repr__(self):
        return "\n".join([str(r) for r in self.rules])


class JoinSequence(object):
    """Convenience class for operating on a sequence of joins.
  A planner takes a joingraph and emits a join sequence.
  It's not yet a relational plan; we still need to do some bookkeeping
  with respect to datalog terms"""

    def __init__(self):
        self.terms = []
        self.conditions = []
        self.num_atoms = 0

    def offset(self, term):
        """Find the position offset for columns in term.
    For example, offset([A(X,Y), B(Y,Z)], B) == 2
    and offset([A(X,Y), B(Y,Z)], A) == 0"""
        pos = 0
        for t in self.terms:
            if t == term:
                return pos
            pos = pos + len(t.valuerefs)

        raise KeyError("%s not found in join sequence %s" % (term, self.terms))

    def makePlan(self, selection_conditions, program):
        """Make a relational plan. The selection_conditions are explicit
        conditions from the Datalog rule, like X=3"""
        leaves = [t.makeLeaf(selection_conditions, program)
                  for t in self.terms]

        if not leaves:
            return algebra.EmptyRelation(Scheme())

        leftmost = leaves[0]
        pairs = zip(self.conditions, leaves[1:])

        def makejoin(leftplan, (condition, right)):
            return algebra.Join(condition, leftplan, right)

        return reduce(makejoin, pairs, leftmost)

    def addjoin(self, left, right, condition):
        """Append a join to the sequence.  The left term should already appear
        somewhere in the sequence."""
        if not self.terms:
            self.terms.append(left)
            self.num_atoms += len(left)

        termsToOffset = {left: self.offset(left), right: self.num_atoms}
        LOG.debug("addjoin: before add offset %s", condition)
        condition.add_offset_by_terms(termsToOffset)
        LOG.debug("addjoin: after add offset %s", condition)

        self.terms.append(right)
        self.num_atoms += len(right)
        self.conditions.append(condition)

    def __repr__(self):
        if not self.terms:
            return "EmptyJoinSequence"
        else:
            return " ".join(["%s *%s " % (t, c)
                             for t, c in zip(self.terms, self.conditions)]) \
                + "%s" % self.terms[-1]

    def __iter__(self):
        return self.terms.__iter__()


class Planner(object):
    """Given a join graph, produce a join sequence by choosing a join order
    with respect to some cost function.  Subclasses can implement this however
    they want."""

    def __init__(self, joingraph):
        self.joingraph = joingraph

    def joininfo(self, joinedge):
        """Given a join edge, return a pair of terms and the metadata for the
        join"""
        # TODO: Does this belong here, or in JoinSequence?  Especially the flip
        # logic.
        leftterm, rightterm = joinedge
        data = self.joingraph.get_edge_data(*joinedge)
        condition = data["condition"]

        LOG.debug("joininfo: condition(%s), data(%s), leftterm(%s), rightterm(%s)", condition, data, leftterm, rightterm)  # noqa

        return leftterm, rightterm, condition

    def toJoinSequence(self, edgesequence, joinsequence=None):
        """Convert a sequence of join edges to a JoinSequence object.  A
        joinsequence has the correct offsets and orientation for positional
        references.  An edgesequence does not."""

        if joinsequence is None:
            joinsequence = JoinSequence()

        # TODO: Move this to JoinSequence?
        if not edgesequence:
            return joinsequence

        edge, rem = edgesequence[0], edgesequence[1:]

        left, right, condition = self.joininfo(edge)
        # FIXME: Sometimes condition gets switched here!!

        LOG.debug("toJoinSequence: left(%s), right(%s), conditions(%s) => (calculating)", left, right, condition)  # noqa
        joinsequence.addjoin(left, right, condition)
        LOG.debug("toJoinSequence: left(%s), right(%s), conditions(%s) => joinsequence(%s)", left, right, condition, joinsequence)  # noqa

        js = self.toJoinSequence(rem, joinsequence)
        return js


def normalize(x, y):
    if y < x:
        edge = (y, x)
    else:
        edge = (x, y)
    return edge


class BFSLeftDeepPlanner(Planner):

    def chooseplan(self, costfunc=None):
        """Return a join sequence object based on the join graph.  This one is
        simple -- it just adds the joins according to a breadth first search"""
        # choose first node in the original insertion order (networkx does not
        # guarantee even a deterministic order)
        firstnode = [n for n in self.joingraph.nodes()
                     if n.originalorder == 0][0]

        # get a BFS ordering of the edges.  Ignores costs.
        edgesequence = [x for x in nx.bfs_edges(self.joingraph, firstnode)]

        LOG.debug("BFS: edgesequence: %s", edgesequence)

        # Make it deterministic but still in BFS order
        deterministic_edge_sequence = []
        while len(edgesequence) > 0:
            # Consider all edges that have the same first node -- these are all
            # "ties" in BFS order.
            firstx = edgesequence[0][0]
            new_edges = [(x, y) for (x, y) in edgesequence if x == firstx]
            # Sort edges on the originalorder of the source and destination
            deterministic_edge_sequence.extend(sorted(new_edges, key=lambda (x, y): (x.originalorder, y.originalorder)))  # noqa
            # Remove all those edges from edgesequence
            edgesequence = [(x, y) for (x, y) in edgesequence if x != firstx]

        LOG.debug("BFS: deterministic edge seq: %s",
                  deterministic_edge_sequence)

        # Generate a concrete sequence of terms with conditions properly
        # adjusted
        joinsequence = self.toJoinSequence(deterministic_edge_sequence)
        LOG.debug("BFS: joinsequence: %s", joinsequence)
        return joinsequence


class Rule(object):
    def __init__(self, headbody):
        self.head = headbody[0]
        self.body = headbody[1]
        # flag used to detect recursion
        self.compiling = False
        self.fixpoint = None

    def vars(self):
        """Return a list of variables in their order of appearence in the rule.
        No attempt to remove duplicates"""
        for term in self.body:
            for v in term.vars():
                yield v

    def refersTo(self, term):
        """Return true if this rule includeas a reference to the given term"""
        for bodyterm in self.body:
            if isinstance(bodyterm, Term) and bodyterm.samerelation(term):
                return True
        return False

    def isParallel(self):
        if self.head.serverspec:
            return True
        return False

    def IDBof(self, term):
        """Return true if this rule defines an IDB corresponding to the given
        term"""
        return term.name == self.head.name and \
            len(term.valuerefs) == len(self.head.valuerefs)

    def toRA(self, program):
        """Emit a relational plan for this rule"""
        if program.compiling(self.head):
            # recursive rule
            if not self.fixpoint:
                self.fixpoint = algebra.Fixpoint()
            state = algebra.State(self.head.name, self.fixpoint)
            return state
        else:
            self.compiling = True

        # get the terms, like A(X,Y,"foo")
        terms = [c for c in self.body if isinstance(c, Term)]

        # get the conditions, like Z=3
        conditions = [c for c in self.body
                      if isinstance(c, expression.BinaryBooleanOperator)]
        if len(conditions) > 0:
            LOG.debug("found conditions: %s (type=%s) for program %s", conditions, type(conditions[0]), program)  # noqa
        else:
            LOG.debug("found conditions: %s (type=%s) for program %s", conditions, None, program)  # noqa

        # construct the join graph
        joingraph = nx.Graph()
        N = len(terms)
        for i, term1 in enumerate(terms):
            # store the order for explaining queries later -- not strictly
            # necessary
            term1.originalorder = i

            # for each term, add it as a vertex,
            # and for each term it joins to, add an edge
            joingraph.add_node(term1, term=term1)
            for j in range(i + 1, N):
                term2 = terms[j]
                LOG.debug("joinsto? %s %s", term1, term2)
                joins = term1.joinsto(term2, conditions)
                if joins:
                    conjunction = reduce(expression.AND, joins)
                    LOG.debug("add edge: %s --[%s]--> %s", term1, conjunction,
                              term2)
                    joingraph.add_edge(term1, term2, condition=conjunction,
                                       terms=(term1, term2))

        # find connected components (some non-determinism in the order here)
        comps = nx.connected_component_subgraphs(joingraph)

        component_plans = []

        # for each component, choose a join order
        for component in comps:
            cycleconditions = []
            # check for cycles
            cycles = nx.cycle_basis(component)
            while cycles:
                LOG.debug("found cycles: %s", cycles)

                # choose an edge to break the cycle
                # that edge will be a selection condition after the final join
                # oneedge = cycles[0][-2:]
                # try to make the chosen edge from cycle deterministic
                oneedge = sorted(cycles[0], key=lambda v: v.originalorder)[-2:]

                data = component.get_edge_data(*oneedge)
                LOG.debug("picked edge: %s, data: %s", oneedge, data)
                cycleconditions.append(data)
                component.remove_edge(*oneedge)
                cycles = nx.cycle_basis(component)

            if len(component) == 1:
                # no joins to plan
                onlyterm = component.nodes()[0]
                plan = onlyterm.makeLeaf(conditions, program)
            else:
                LOG.debug("component: %s", component)
                # TODO: clean this up.
                # joingraph -> joinsequence -> relational plan
                planner = BFSLeftDeepPlanner(component)

                joinsequence = planner.chooseplan()
                LOG.debug("join sequence: %s", joinsequence)

                # create a relational plan, finally
                # pass in the conditions to make the leaves of the plan
                plan = joinsequence.makePlan(conditions, program)

            LOG.debug("cycleconditions: %s", cycleconditions)
            for condition_info in cycleconditions:
                predicate = condition_info["condition"]
                terms = condition_info["terms"]

                # change all UnnamedAttributes based on the
                # offset of its Term
                termsToOffset = dict((t, joinsequence.offset(t))
                                     for t in terms)

                LOG.debug("before add offset %s", predicate)
                predicate.add_offset_by_terms(termsToOffset)
                LOG.debug("after add offset %s", predicate)

                # create selections after each cycle
                plan = algebra.Select(predicate, plan)

            component_plans.append(plan)

        # link the components with a cross product
        plan = component_plans[0]
        for newplan in component_plans[1:]:
            plan = algebra.CrossProduct(plan, newplan)

        try:
            scheme = plan.scheme()
        except AttributeError:
            scheme = Scheme([make_attr(i, r, self.head.name) for i, r in enumerate(self.head.valuerefs)])  # noqa

        # Helper function for the next two steps (TODO: move this to a method?)
        def findvar(variable):
            var = variable.var
            if var not in scheme:
                msg = "Head variable %s does not appear in rule body: %s" % (var, self)  # noqa
                raise SyntaxError(msg)
            return expression.UnnamedAttributeRef(scheme.getPosition(var))

        class FindVarExpressionVisitor(expression.SimpleExpressionVisitor):
            def __init__(self):
                self.stack = []

            def getresult(self):
                assert len(self.stack) == 1
                return self.stack.pop()

            def visit_unary(self, unaryexpr):
                inputexpr = self.stack.pop()
                self.stack.append(unaryexpr.__class__(inputexpr))

            def visit_binary(self, binaryexpr):
                right = self.stack.pop()
                left = self.stack.pop()
                self.stack.append(binaryexpr.__class__(left, right))

            def visit_zeroary(self, zeroaryexpr):
                self.stack.append(zeroaryexpr.__class__())

            def visit_literal(self, literalexpr):
                self.stack.append(literalexpr.__class__(literalexpr.value))

            def visit_nary(self, naryexpr):
                raise NotImplementedError(
                    "TODO: implement findvar visit of nary expression")

            def visit_Var(self, var):
                asAttr = findvar(var)
                self.stack.append(asAttr)

            # TODO: add the other aggregates
            # TODO and move aggregates to expression-visitor
            def visit_SUM(self, x):
                self.visit_unary(x)

            def visit_COUNT(self, x):
                self.visit_unary(x)

        # if this Rule includes a server specification, add a partition
        # operator
        if self.isParallel():
            if isinstance(self.head.serverspec, Broadcast):
                plan = algebra.Broadcast(plan)
            if isinstance(self.head.serverspec, PartitionBy):
                positions = [findvar(v)
                             for v in self.head.serverspec.variables]
                plan = algebra.PartitionBy(positions, plan)

        def toAttrRef(e):
            """
             Resolve variable references in the head; pass through aggregate
             expressions

             If expression requires an Apply then return True, else False
             """
            LOG.debug("find reference for %s", e)
            visitor = FindVarExpressionVisitor()
            e.accept(visitor)
            return visitor.getresult()

        columnlist = [toAttrRef(v) for v in self.head.valuerefs]
        LOG.debug("columnlist for Project (or group by) is %s", columnlist)

        # If any of the expressions in the head are aggregate expression,
        # construct a group by
        if any(expression.expression_contains_aggregate(v)
               for v in self.head.valuerefs):
            emit_clause = [(None, a_or_g) for a_or_g in columnlist]
            return raco.myrial.groupby.groupby(plan, emit_clause, [])
        elif any([not isinstance(e, Var) for e in self.head.valuerefs]):
            # If complex expressions in head, then precede Project with Apply
            # NOTE: should Apply actually just append emitters to schema
            # instead of doing column select?
            # we decided probably not in
            # https://github.com/uwescience/raco/pull/209
            plan = algebra.Apply([(None, e) for e in columnlist], plan)
        else:
            # otherwise, just build a Project
            plan = algebra.Apply(emitters=[(None, c) for c in columnlist],
                                 input=plan)

        # If we found a cycle, the "root" of the plan is the fixpoint operator
        if self.fixpoint:
            self.fixpoint.loopBody(plan)
            plan = self.fixpoint
            self.fixpoint = None

        self.compiling = False

        return plan

    def __repr__(self):
        return "%s :- %s" % (self.head, ", ".join([str(t) for t in self.body]))


class Var(expression.Expression):
    def __init__(self, var):
        self.var = var

    def vars(self):
        """Works with BinaryBooleanOperator.vars to return a list of vars from
        any expression"""
        return [self.var]

    def __str__(self):
        return str(self.var)

    def __repr__(self):
        return str(self.var)

    def evaluate(self, _tuple, scheme, state=None):
        raise NotImplementedError()

    def apply(self, f):
        raise NotImplementedError()

    def typeof(self, scheme, state_scheme):
        # WRONG: we should read this from a catalogue
        return raco.types.LONG_TYPE

    def get_children(self):
        return []


class Term(object):
    def __init__(self, parsedterm):
        self.name = parsedterm[0]
        self.alias = self.name
        self.valuerefs = [vr for vr in parsedterm[1]]
        self.originalorder = None

    def __len__(self):
        """How many atoms are in this term"""
        return len(self.valuerefs)

    def setalias(self, alias):
        """Assign an alias for this term. Used when the same relation appears
        twice in one rule."""
        self.alias = alias

    def samerelation(self, term):
        """Return True if the argument refers to the same relation"""
        return self.name == term.name

    def __repr__(self):
        return "%s_%s(%s)" % (self.__class__.__name__, self.name,
                              ",".join([str(e) for e in self.valuerefs]))

    def vars(self):
        """Return a list of variable names used in this term."""
        return [vr.var for vr in self.valuerefs if isinstance(vr, Var)]

    def varpos(self):
        """Return a list of (position, variable name) tuples for variables used
        in this term."""
        return [(pos, vr.var)
                for (pos, vr) in enumerate(self.valuerefs)
                if isinstance(vr, Var)]

    def labelAttr(self, attr):
        LOG.debug("label %s with term %s", attr, self)
        # store a reference to the term this AttributeRef comes from
        attr.myTerm = self

    def joinsto(self, other, conditions):
        """Return the join conditions between this term and the argument term.
        The second argument is a list of explicit conditions, like X=3 or Y=Z

        The attributes for the variables in the conditions will be labeled with
        the term they are from."""

        # get the implicit join conditions
        yourvars = other.varpos()
        myvars = self.varpos()
        varjoins = []
        Pos = expression.UnnamedAttributeRef
        for i, var in myvars:
            match = [j for (j, var2) in yourvars if var2 == var]
            if match:
                myposition = Pos(i)
                self.labelAttr(myposition)
                # TODO this code only picks the first matching variable.
                # Consider:
                #    A(x) :- R(x), S(x,x). We will end up with
                #
                #    Project($0)[Join($0=$0)[Scan(R),Select($0=$1)]Scan(S)]
                #
                # Might it be better to emit both Join conditions instead of
                # the select? In addition to the select?? Might it be better to
                # Join on the second attribute instead? TODO
                yourposition = Pos(match[0])
                other.labelAttr(yourposition)
                varjoins.append(expression.EQ(myposition, yourposition))

        LOG.debug("variable joins: %s", varjoins)

        condjoins = []

        # get the explicit join conditions
        # TODO: this only works for BinaryOperators of depth 1 (binop left
        # right). It can be generalized
        for c in conditions:
            if isinstance(c.left, Var) and isinstance(c.right, Var):
                # then we have a potential join condition
                if self.match(c.left) and other.match(c.right):
                    LOG.debug("match condition %s: self(%s) matches %s; other(%s) matches %s", c, self, c.left, other, c.right)  # noqa
                    leftAttr = self.convertvalref(c.left)
                    rightAttr = other.convertvalref(c.right)
                    self.labelAttr(leftAttr)
                    other.labelAttr(rightAttr)
                    condjoins.append(c.__class__(leftAttr, rightAttr))
                elif other.match(c.left) and self.match(c.right):
                    LOG.debug("match condition %s: self(%s) matches %s; other(%s) matches %s", c, self, c.right, other, c.left)  # noqa
                    leftAttr = other.convertvalref(c.left)
                    rightAttr = self.convertvalref(c.right)
                    self.labelAttr(rightAttr)
                    other.labelAttr(leftAttr)
                    condjoins.append(c.__class__(leftAttr, rightAttr))
                else:
                    # must be a condition on some other pair of relations
                    pass

        LOG.debug("cond joins: %s", condjoins)

        return varjoins + condjoins

    def match(self, valref):
        """Return true if valref is a variable and is used in the list"""
        return isinstance(valref, Var) and valref.var in self.vars()

    def position(self, valueref):
        """ Returns the position of the variable in the term.  Throws an error
        if it does not appear."""
        LOG.debug("position: vars(%s), valueref.var(%s)", [v for v in self.valuerefs], valueref.var)  # noqa
        return dict([(var, pos) for pos, var in self.varpos()])[valueref.var]

    def convertvalref(self, valueref):
        """Convert a Datalog value reference (a literal or a variable) to RA.
        Literals are passed through unchanged.  Variables are converted"""
        if self.match(valueref):
            pos = self.position(valueref)
            LOG.debug("convertvalref(match) self(%s), valueref(%s), pos(%s)", self, valueref, pos)  # noqa

            return expression.UnnamedAttributeRef(pos)
        else:
            # must be a join condition, and the other variable matches
            LOG.debug("convertvalref(not) self(%s), valueref(%s)", self, valueref)  # noqa
            return valueref

    def explicitconditions(self, conditions):
        """Given a list of parsed Datalog boolean conditions, return an
        iterator over RA selection conditions that apply to this term.  Ignore
        join conditions."""
        # This method is written rather verbosely for clarity

        Literal = expression.Literal

        for condition in conditions:
            cons = condition.__class__

            if isinstance(condition.left, Literal) and \
                    isinstance(condition.right, Literal):
                # No variables
                yield cons(condition.left, condition.right)

            elif isinstance(condition.left, Literal) and \
                    isinstance(condition.right, Var):
                # left literal, right variable, like 3 = X
                if self.match(condition.right):
                    yield cons(condition.left,
                               self.convertvalref(condition.right))
                else:
                    # variable is not used in this term, so do nothing
                    pass

            elif isinstance(condition.left, Var) and \
                    isinstance(condition.right, Literal):
                # left variable, right literal, like X = 3
                if self.match(condition.left):
                    yield cons(self.convertvalref(condition.left),
                               condition.right)
                else:
                    # variable is not used in this term, so do nothing
                    pass

            elif isinstance(condition.left, Var) and \
                    isinstance(condition.right, Var):
                # two variables. Both match, neither match, or one matches.
                leftmatch = self.match(condition.left)
                rightmatch = self.match(condition.right)
                if leftmatch and rightmatch:
                    # a selection condition like A(X,Y), X=Y
                    yield cons(self.convertvalref(condition.left),
                               self.convertvalref(condition.right))
                elif not leftmatch and not rightmatch:
                    # nothing to do with this term
                    pass
                elif not leftmatch and rightmatch:
                    # join condition, leave it alone
                    pass
                elif leftmatch and not rightmatch:
                    # join condition, leave it alone
                    pass

    def implicitconditions(self):
        """An iterator over implicit selection conditions derived from the
        datalog syntax. For example, A(X,X) implies position0 == position1,
        and A(X,4) implies position1 == 4"""
        # Check for implicit literal equality conditions, like A(X,"foo")
        for i, b in enumerate(self.valuerefs):
            if isinstance(b, expression.Literal):
                posref = expression.UnnamedAttributeRef(i)
                yield expression.EQ(posref, b)

        # Check for repeated variable conditions, like A(X,X)
        N = len(self.valuerefs)
        for i, x in enumerate(self.valuerefs):
            if isinstance(x, Var):
                for j in range(i + 1, N):
                    y = self.valuerefs[j]
                    if isinstance(y, Var):
                        # TODO: probably want to implement __eq__, but it makes
                        # Var objects unhashable
                        if x.var == y.var:
                            leftpos = expression.UnnamedAttributeRef(i)
                            rightpos = expression.UnnamedAttributeRef(j)
                            yield expression.EQ(leftpos, rightpos)

    def renameIDB(self, plan):
        """Rename the attributes of the plan to match the current rule. Used
        when chaining multiple rules."""
        term = self

        # Do we know the actual scheme? If it's recursive, we don't
        # So derive it from the rule head
        # Not really satisfied with this.
        try:
            sch = plan.scheme()
        except algebra.RecursionError:
            sch = Scheme([make_attr(i, r, term.name)
                          for i, r in enumerate(term.valuerefs)])

        oldscheme = [name for (name, _) in sch]
        termscheme = [expr for expr in term.valuerefs]

        if len(oldscheme) != len(termscheme):
            raise TypeError("Rule with head %s does not match Term %s" % (term.name, term))  # noqa

        pairs = zip(termscheme, oldscheme)

        # Merge the old and new schemes.  Use new where we can.
        def choosename(new, old):
            if isinstance(new, Var):
                # Then use this new var as the column name
                return new.var
            else:
                # It's an implicit selection condition, like R(x,3). In R's
                # schema, don't call the column name '3' (new), instead call it
                # whatever the column name was in the prior rule where R is the
                # head variable R (old).
                return old

        mappings = [(choosename(new, old), expression.UnnamedAttributeRef(i))
                    for i, (new, old) in enumerate(pairs)]

        # Use an apply operator to implement the renaming
        plan = algebra.Apply(mappings, plan)

        return plan

    def makeLeaf(self, conditions, program):
        """Return an RA plan that Scans the appropriate relation and applies
        all selection conditions. Two sources of conditions: the term itself,
        like A(X,"foo") -> Select(pos1="foo", Scan(A)) and separate condition
        terms, like A(X,Y), X=3 -> Select(pos0=3, Scan(A)) separate condition
        terms are passed in as an argument."""

        # Chain rules together
        if program.isIDB(self):
            plan = program.compileIDB(self.name)
            scan = self.renameIDB(plan)
        else:
            sch = Scheme([make_attr(i, r, self.name)
                          for i, r in enumerate(self.valuerefs)])
            rel_key = RelationKey.from_string(self.name)
            scan = algebra.Scan(rel_key, sch)
            scan.trace("originalterm", "%s (position %s)" % (self, self.originalorder))  # noqa

        # collect conditions within the term itself, like A(X,3) or A(Y,Y)
        implconds = list(self.implicitconditions())

        # collect explicit conditions, like A(X,Y), Y=2
        explconds = list(self.explicitconditions(conditions))

        allconditions = implconds + explconds

        if allconditions:
            conjunction = reduce(expression.AND, allconditions)
            plan = algebra.Select(conjunction, scan)
        else:
            plan = scan

        plan.set_alias(self)
        return plan


class IDB(Term):
    def __init__(self, termobj, serverspec=None, timestep=None):
        if timestep:
            self.name = "%s_%s" % (termobj.name, timestep.spec)
        else:
            self.name = termobj.name
        self.valuerefs = termobj.valuerefs
        self.serverspec = serverspec
        self.timestep = timestep


class EDB(Term):
    pass


class ServerSpecification(object):
    pass


class Broadcast(ServerSpecification):
    pass


class PartitionBy(ServerSpecification):
    def __init__(self, variables):
        self.variables = variables


class Timestep(object):
    def __init__(self, spec):
        self.spec = spec
        pass



from collections import Counter
import sqlalchemy
import unittest

import raco.algebra as algebra
from raco.compile import optimize_by_rules
from raco.language.logical import OptLogicalAlgebra
import raco.myrial.interpreter as interpreter
import raco.myrial.parser as parser
import raco.scheme as scheme
from .catalog import SQLCatalog
import raco.types as types


class SQLTestCase(unittest.TestCase):
    """A base for testing the compilation of RACO programs to SQL queries"""

    emp_table = [
        # id dept_id name salary
        (1, 2, "Bill Howe", 25000),
        (2, 1, "Dan Halperin", 90000),
        (3, 1, "Andrew Whitaker", 5000),
        (4, 2, "Shumo Chu", 5000),
        (5, 1, "Victor Almeida", 25000),
        (6, 3, "Dan Suciu", 90000),
        (7, 1, "Magdalena Balazinska", 25000)]

    emp_schema = scheme.Scheme([("id", types.INT_TYPE),
                                ("dept_id", types.INT_TYPE),
                                ("name", types.STRING_TYPE),
                                ("salary", types.LONG_TYPE)])

    emp_key = "public:adhoc:employee"

    def setUp(self):
        # SQLAlchemy
        self.db = SQLCatalog(sqlalchemy.
                             create_engine('sqlite:///:memory:', echo=True))
        self.db.add_table(self.emp_key, self.emp_schema)
        self.db.add_tuples(self.emp_key, self.emp_schema, self.emp_table)
        # MyriaL
        self.parser = parser.Parser()
        self.processor = interpreter.StatementProcessor(self.db)

    def query_to_phys_plan(self, query, **kwargs):
        statements = self.parser.parse(query)
        self.processor.evaluate(statements)
        p = self.processor.get_logical_plan(**kwargs)
        p = optimize_by_rules(p, OptLogicalAlgebra.opt_rules())
        if isinstance(p, (algebra.Store, algebra.StoreTemp)):
            p = p.input
        return p

    def execute(self, query, expected, **kwargs):
        p = self.query_to_phys_plan(query, **kwargs)
        ans = self.db.evaluate(p)
        self.assertEquals(expected, Counter(ans))


"""Test of aggregations over empty relations.

Aggregation queries without grouping should return sensible default values:
COUNT(empty) == 0
SUM(empty) == 0
"""

import collections

import raco.myrial.myrial_test as myrial_test


class EmptyAggregateTests(myrial_test.MyrialTestCase):

    def setUp(self):
        super(EmptyAggregateTests, self).setUp()

    def test_count(self):
        query = """
        W = EMPTY(v:int);
        X = [FROM W EMIT COUNT(v)];
        STORE(X, OUTPUT);
        """

        self.check_result(query, collections.Counter([(0,)]))

    def test_sum(self):
        query = """
        W = EMPTY(v:int);
        X = [FROM W EMIT SUM(v)];
        STORE(X, OUTPUT);
        """

        self.check_result(query, collections.Counter([(0,)]))




import abc
from raco.utility import emitlist
from algebra import gensym, Operator

import logging
LOG = logging.getLogger(__name__)


class ResolvingSymbol:
    def __init__(self, name):
        self._name = name
        self._placeholder = "%%(%s)s" % name

    def getPlaceholder(self):
        return self._placeholder

    def getName(self):
        return self._name


class CompileState:

    def __init__(self, lang, cse=True):
        self.language = lang

        self.declarations = []
        self.declarations_later = []
        self.pipelines = []
        self.scan_pipelines = []
        self.flush_pipelines = []
        self.initializers = []
        self.pipeline_count = 0

        # { expression => symbol for materialized result }
        self.materialized = {}

        # { symbol => tuple type definition }
        self.tupledefs = {}

        # symbol resolution
        self.resolving_symbols = {}

        self.common_subexpression_elim = cse

        self.current_pipeline_properties = {}
        self.current_pipeline_precode = []
        self.current_pipeline_postcode = []

        self.main_wait_statements = set()

    def setPipelineProperty(self, key, value):
        LOG.debug("set %s in %s" % (key, self.current_pipeline_properties))
        self.current_pipeline_properties[key] = value

    def addToPipelinePropertyList(self, key, value):
        current = self.current_pipeline_properties.get(key, [])
        assert isinstance(current, list), "cannot add to non-list property"
        if len(current) == 0:
            self.current_pipeline_properties[key] = current

        current.append(value)

    def addToPipelinePropertySet(self, key, value):
        current = self.current_pipeline_properties.get(key, set())
        assert isinstance(current, set), "cannot add to non-set property"
        if len(current) == 0:
            self.current_pipeline_properties[key] = current

        current.add(value)

    def getPipelineProperty(self, key):
        LOG.debug("get %s from %s" % (key, self.current_pipeline_properties))
        return self.current_pipeline_properties[key]

    def checkPipelineProperty(self, key):
        """
        Like getPipelineProperty but returns None if no property is found
        """
        LOG.debug("get(to check) %s from %s"
                  % (key, self.current_pipeline_properties))
        return self.current_pipeline_properties.get(key)

    def createUnresolvedSymbol(self):
        name = gensym()
        rs = ResolvingSymbol(name)
        self.resolving_symbols[name] = None
        return rs

    def resolveSymbol(self, rs, value):
        self.resolving_symbols[rs.getName()] = value

    def addDeclarations(self, d):
        self.declarations += d

    def addDeclarationsUnresolved(self, d):
        """
        Ordered in the code after the regular declarations
        just so that any name dependences already have been declared
        ALTERNATIVE: split decls into forward decls and definitions
        """
        self.declarations_later += d

    def addInitializers(self, i):
        self.initializers += i

    def addMainWaitStatement(self, c):
        self.main_wait_statements.add(c)

    def addPipeline(self, p):
        LOG.debug("output pipeline %s", self.current_pipeline_properties)
        pipeline_code = \
            emitlist(self.current_pipeline_precode) +\
            self.language.pipeline_wrap(self.pipeline_count, p,
                                        self.current_pipeline_properties) +\
            emitlist(self.current_pipeline_postcode)

        # force scan pipelines to go first
        if self.current_pipeline_properties.get('type') == 'scan':
            self.scan_pipelines.append(pipeline_code)
        else:
            self.pipelines.append(pipeline_code)

        self.pipeline_count += 1
        self.current_pipeline_properties = {}
        self.current_pipeline_precode = []
        self.current_pipeline_postcode = []

    def addCode(self, c):
        """
        Just add code here
        """
        self.pipelines.append(c)

    def addPreCode(self, c):
        self.current_pipeline_precode.append(c)

    def addPostCode(self, c):
        self.current_pipeline_postcode.append(c)

    def addPipelineFlushCode(self, c):
        self.flush_pipelines.append(c)

    def getInitCode(self):

        # inits is a set.
        # If this ever becomes a bottleneck when declarations are strings,
        # as in clang, then resort to at least symbol name deduping.
        # TODO: better would be to mark elements of self.initializers as
        # TODO: "do dedup" or "don't dedup"
        s = set()

        def f(x):
            if x in s:
                return False
            else:
                s.add(x)
                return True

        code = emitlist(filter(f, self.initializers))
        return code % self.resolving_symbols

    def getDeclCode(self):
        # declarations is a set
        # If this ever becomes a bottleneck when declarations are strings,
        # as in clang, then resort to at least symbol name deduping.
        s = set()

        def f(x):
            if x in s:
                return False
            else:
                s.add(x)
                return True

        # keep in original order
        code = emitlist(filter(f, self.declarations))
        code += emitlist(filter(f, self.declarations_later))
        return code % self.resolving_symbols

    def getExecutionCode(self):
        # list -> string
        scan_linearized = emitlist(self.scan_pipelines)
        mem_linearized = \
            emitlist(self.pipelines) + emitlist(self.main_wait_statements)
        flush_linearized = emitlist(self.flush_pipelines)
        scan_linearize_wrap = self.language.group_wrap(gensym(),
                                                       scan_linearized,
                                                       {'type': 'scan'})
        mem_linearize_wrap = self.language.group_wrap(gensym(),
                                                      mem_linearized,
                                                      {'type': 'in_memory'})

        linearized = \
            scan_linearize_wrap + mem_linearize_wrap + flush_linearized

        # substitute all lazily resolved symbols
        resolved = linearized % self.resolving_symbols

        return resolved

    def lookupExpr(self, expr):

        if self.common_subexpression_elim:
            res = self.materialized.get(expr)
            LOG.debug("lookup subexpression %s -> %s", expr, res)
            return res
        else:
            # if CSE is turned off
            # then always return None for expression matches
            return None

    def saveExpr(self, expr, sym):
        LOG.debug("saving subexpression %s -> %s", expr, sym)
        self.materialized[expr] = sym

    def lookupTupleDef(self, sym):
        return self.tupledefs.get(sym)

    def saveTupleDef(self, sym, tupledef):
        self.tupledefs[sym] = tupledef

    def getCurrentPipelineId(self):
        return self.pipeline_count


class Pipelined(object):
    """
    Trait to provide the compilePipeline method
    for calling into pipeline style compilation.

    This is a mixin class that supports cooperative
    multiple inheritance. To use it properly, put
    it _before_ the base class in the inheritance clause
    """

    __metaclass__ = abc.ABCMeta

    def __init__(self, *args):
        self._parent = None
        # Ensure this class follows cooperative multiple inheritance
        super(Pipelined, self).__init__(*args)

    def __markAllParents__(self):
        root = self

        def markChildParent(op):
            for c in op.children():
                c._parent = op
            return []

        [_ for _ in root.postorder_traversal(markChildParent)]

    def parent(self):
        return self._parent

    @classmethod
    @abc.abstractmethod
    def language(cls):
        pass

    @classmethod
    @abc.abstractmethod
    def new_tuple_ref(cls, symbol, scheme):
        pass

    @abc.abstractmethod
    def postorder_traversal(self, func):
        pass

    @abc.abstractmethod
    def produce(self, state):
        """Denotation for producing a tuple"""
        return

    @abc.abstractmethod
    def consume(self, inputTuple, fromOp, state):
        """Denotation for consuming a tuple"""
        return

    def compilePipeline(self):
        self.__markAllParents__()

        state = CompileState(self.language())

        state.addCode(
            self.language().comment("Compiled subplan for %s" % self))

        self.produce(state)

        # state.addCode( self.language().log("Evaluating subplan %s" % self) )

        return state



from setuptools import setup

### Add find_packages function, see
# https://wiki.python.org/moin/Distutils/Cookbook/AutoPackageDiscovery
import os

def is_package(path):
    return (
        os.path.isdir(path) and
        os.path.isfile(os.path.join(path, '__init__.py'))
        )

def find_packages(path=".", base="", exclude=None):
    """Find all packages in path"""
    if not exclude:
        exclude = []
    packages = {}
    for item in os.listdir(path):
        dir = os.path.join(path, item)
        if is_package(dir) and dir not in exclude:
            if base:
                module_name = "{base}.{item}".format(base=base,item=item)
            else:
                module_name = item
            packages[module_name] = dir
            packages.update(find_packages(dir, module_name))
    return packages
###

setup(name='raco',
      version='1.1.0',
      description='Relational Algebra COmpiler',
      author='Bill Howe, Andrew Whitaker, Daniel Halperin',
      author_email='raco@cs.washington.edu',
      url='https://github.com/uwescience/raco',
      packages=find_packages(exclude=['clang']),
      package_data={'': ['c_templates/*.template','grappa_templates/*.template']},
      install_requires=['networkx', 'ply', 'pyparsing', 'SQLAlchemy'],
      scripts=['scripts/myrial']
      )



import collections
import random

from raco.algebra import *
from raco.expression import NamedAttributeRef as AttRef
from raco.expression import UnnamedAttributeRef as AttIndex
from raco.expression import StateVar

from raco.language.myrialang import (
    MyriaShuffleConsumer, MyriaShuffleProducer, MyriaHyperShuffleProducer,
    MyriaBroadcastConsumer, MyriaQueryScan)
from raco.language.myrialang import (MyriaLeftDeepTreeAlgebra,
                                     MyriaHyperCubeAlgebra)
from raco.compile import optimize
from raco import relation_key
from raco.catalog import FakeCatalog

import raco.scheme as scheme
import raco.myrial.myrial_test as myrial_test
from raco import types


class OptimizerTest(myrial_test.MyrialTestCase):

    x_scheme = scheme.Scheme([("a", types.LONG_TYPE), ("b", types.LONG_TYPE), ("c", types.LONG_TYPE)])  # noqa
    y_scheme = scheme.Scheme([("d", types.LONG_TYPE), ("e", types.LONG_TYPE), ("f", types.LONG_TYPE)])  # noqa
    x_key = relation_key.RelationKey.from_string("public:adhoc:X")
    y_key = relation_key.RelationKey.from_string("public:adhoc:Y")

    def setUp(self):
        super(OptimizerTest, self).setUp()

        random.seed(387)  # make results deterministic
        rng = 20
        count = 30
        self.x_data = collections.Counter(
            [(random.randrange(rng), random.randrange(rng),
              random.randrange(rng)) for _ in range(count)])
        self.y_data = collections.Counter(
            [(random.randrange(rng), random.randrange(rng),
              random.randrange(rng)) for _ in range(count)])

        self.db.ingest(OptimizerTest.x_key,
                       self.x_data,
                       OptimizerTest.x_scheme)
        self.db.ingest(OptimizerTest.y_key,
                       self.y_data,
                       OptimizerTest.y_scheme)

        self.expected = collections.Counter(
            [(a, b, c, d, e, f) for (a, b, c) in self.x_data
             for (d, e, f) in self.y_data if a > b and e <= f and c == d])

        self.z_key = relation_key.RelationKey.from_string("public:adhoc:Z")
        self.z_data = collections.Counter([(1, 2), (2, 3), (1, 2), (3, 4)])
        self.z_scheme = scheme.Scheme([('src', types.LONG_TYPE), ('dst', types.LONG_TYPE)])  # noqa
        self.db.ingest('public:adhoc:Z', self.z_data, self.z_scheme)

        self.expected2 = collections.Counter(
            [(s1, d3) for (s1, d1) in self.z_data.elements()
             for (s2, d2) in self.z_data.elements()
             for (s3, d3) in self.z_data.elements() if d1 == s2 and d2 == s3])

    @staticmethod
    def logical_to_physical(lp, **kwargs):
        if kwargs.get('hypercube', False):
            algebra = MyriaHyperCubeAlgebra(FakeCatalog(64))
        else:
            algebra = MyriaLeftDeepTreeAlgebra()
        return optimize(lp, algebra, **kwargs)

    @staticmethod
    def get_count(op, claz):
        """Return the count of operator instances within an operator tree."""

        def count(_op):
            if isinstance(_op, claz):
                yield 1
            else:
                yield 0
        return sum(op.postorder(count))

    @staticmethod
    def get_num_select_conjuncs(op):
        """Get the number of conjunctions within all select operations."""
        def count(_op):
            if isinstance(_op, Select):
                yield len(expression.extract_conjuncs(_op.condition))
            else:
                yield 0
        return sum(op.postorder(count))

    def test_push_selects(self):
        """Test pushing selections into and across cross-products."""
        lp = StoreTemp('OUTPUT',
               Select(expression.LTEQ(AttRef("e"), AttRef("f")),
                 Select(expression.EQ(AttRef("c"), AttRef("d")),
                   Select(expression.GT(AttRef("a"), AttRef("b")),
                      CrossProduct(Scan(self.x_key, self.x_scheme),
                                   Scan(self.y_key, self.y_scheme))))))  # noqa

        self.assertEquals(self.get_count(lp, Select), 3)
        self.assertEquals(self.get_count(lp, CrossProduct), 1)

        pp = self.logical_to_physical(lp)
        self.assertIsInstance(pp.input, Join)
        self.assertEquals(self.get_count(pp, Select), 2)
        self.assertEquals(self.get_count(pp, CrossProduct), 0)

        self.db.evaluate(pp)
        result = self.db.get_temp_table('OUTPUT')
        self.assertEquals(result, self.expected)

    def test_collapse_applies(self):
        """Test pushing applies together."""
        lp = StoreTemp('OUTPUT',
               Apply([(None, AttIndex(1)), ('w', expression.PLUS(AttIndex(0), AttIndex(0)))],       # noqa
                 Apply([(None, AttIndex(1)), (None, AttIndex(0)), (None, AttIndex(1))],             # noqa
                   Apply([('x', AttIndex(0)), ('y', expression.PLUS(AttIndex(1), AttIndex(0)))],    # noqa
                     Apply([(None, AttIndex(0)), (None, AttIndex(1))],
                           Scan(self.x_key, self.x_scheme))))))  # noqa

        self.assertEquals(self.get_count(lp, Apply), 4)

        pp = self.logical_to_physical(lp)
        self.assertIsInstance(pp.input, Apply)
        self.assertEquals(self.get_count(pp, Apply), 1)

        expected = collections.Counter(
            [(b, a + a) for (a, b, c) in
             [(b, a, b) for (a, b) in
              [(a, b + a) for (a, b) in
                [(a, b) for (a, b, c) in self.x_data]]]])  # noqa
        self.db.evaluate(pp)
        result = self.db.get_temp_table('OUTPUT')
        self.assertEquals(result, expected)

    def test_select_count_star(self):
        """Test that we don't generate 0-length applies from a COUNT(*)."""
        lp = StoreTemp('OUTPUT',
                       GroupBy([], [expression.COUNTALL()],
                               Scan(self.x_key, self.x_scheme)))

        self.assertEquals(self.get_count(lp, GroupBy), 1)

        pp = self.logical_to_physical(lp)
        self.assertIsInstance(pp.input, GroupBy)
        # GroupBy.CollectProducer.CollectConsumer.GroupBy.Apply
        apply = pp.input.input.input.input.input
        self.assertIsInstance(apply, Apply)
        self.assertEquals(self.get_count(pp, Apply), 1)
        self.assertEquals(len(apply.scheme()), 1)

        expected = collections.Counter([(len(self.x_data),)])
        self.db.evaluate(pp)
        result = self.db.get_temp_table('OUTPUT')
        self.assertEquals(result, expected)

    def test_projects_apply_join(self):
        """Test column selection both Apply into ProjectingJoin
        and ProjectingJoin into its input.
        """
        lp = StoreTemp('OUTPUT',
               Apply([(None, AttIndex(1))],
                 ProjectingJoin(expression.EQ(AttIndex(0), AttIndex(3)),
                   Scan(self.x_key, self.x_scheme),
                   Scan(self.x_key, self.x_scheme),
                   [AttIndex(i) for i in xrange(2 * len(self.x_scheme))])))  # noqa

        self.assertIsInstance(lp.input.input, ProjectingJoin)
        self.assertEquals(2 * len(self.x_scheme),
                          len(lp.input.input.scheme()))

        pp = self.logical_to_physical(lp)
        proj_join = pp.input
        self.assertIsInstance(proj_join, ProjectingJoin)
        self.assertEquals(1, len(proj_join.scheme()))
        self.assertEquals(2, len(proj_join.left.scheme()))
        self.assertEquals(1, len(proj_join.right.scheme()))

        expected = collections.Counter(
            [(b,)
             for (a, b, c) in self.x_data
             for (d, e, f) in self.x_data
             if a == d])

        self.db.evaluate(pp)
        result = self.db.get_temp_table('OUTPUT')
        self.assertEquals(result, expected)

    def test_push_selects_apply(self):
        """Test pushing selections through apply."""
        lp = StoreTemp('OUTPUT',
               Select(expression.LTEQ(AttRef("c"), AttRef("a")),
                 Select(expression.LTEQ(AttRef("b"), AttRef("c")),
                   Apply([('b', AttIndex(1)),
                          ('c', AttIndex(2)),
                          ('a', AttIndex(0))],
                         Scan(self.x_key, self.x_scheme)))))  # noqa

        expected = collections.Counter(
            [(b, c, a) for (a, b, c) in self.x_data if c <= a and b <= c])

        self.assertEquals(self.get_count(lp, Select), 2)
        self.assertEquals(self.get_count(lp, Scan), 1)
        self.assertIsInstance(lp.input, Select)

        pp = self.logical_to_physical(lp)
        self.assertIsInstance(pp.input, Apply)
        self.assertEquals(self.get_count(pp, Select), 1)

        self.db.evaluate(pp)
        result = self.db.get_temp_table('OUTPUT')
        self.assertEquals(result, expected)

    def test_push_selects_groupby(self):
        """Test pushing selections through groupby."""
        lp = StoreTemp('OUTPUT',
               Select(expression.LTEQ(AttRef("c"), AttRef("a")),
                 Select(expression.LTEQ(AttRef("b"), AttRef("c")),
                   GroupBy([AttIndex(1), AttIndex(2), AttIndex(0)],
                           [expression.COUNTALL()],
                           Scan(self.x_key, self.x_scheme)))))  # noqa

        expected = collections.Counter(
            [(b, c, a) for (a, b, c) in self.x_data if c <= a and b <= c])
        expected = collections.Counter(k + (v,) for k, v in expected.items())

        self.assertEquals(self.get_count(lp, Select), 2)
        self.assertEquals(self.get_count(lp, Scan), 1)
        self.assertIsInstance(lp.input, Select)

        pp = self.logical_to_physical(lp)
        self.assertIsInstance(pp.input, GroupBy)
        self.assertEquals(self.get_count(pp, Select), 1)

        self.db.evaluate(pp)
        result = self.db.get_temp_table('OUTPUT')
        self.assertEquals(result, expected)

    def test_noop_apply_removed(self):
        lp = StoreTemp('OUTPUT',
               Apply([(None, AttIndex(1))],
                 ProjectingJoin(expression.EQ(AttIndex(0), AttIndex(3)),
                   Scan(self.x_key, self.x_scheme),
                   Scan(self.x_key, self.x_scheme),
                   [AttIndex(i) for i in xrange(2 * len(self.x_scheme))])))  # noqa

        self.assertIsInstance(lp.input, Apply)
        lp_scheme = lp.scheme()

        pp = self.logical_to_physical(lp)
        self.assertNotIsInstance(pp.input, Apply)
        self.assertEquals(lp_scheme, pp.scheme())

    def test_not_noop_apply_not_removed(self):
        lp = StoreTemp('OUTPUT',
               Apply([('hi', AttIndex(1))],
                 ProjectingJoin(expression.EQ(AttIndex(0), AttIndex(3)),
                   Scan(self.x_key, self.x_scheme),
                   Scan(self.x_key, self.x_scheme),
                   [AttIndex(i) for i in xrange(2 * len(self.x_scheme))])))  # noqa

        self.assertIsInstance(lp.input, Apply)
        lp_scheme = lp.scheme()

        pp = self.logical_to_physical(lp)
        self.assertIsInstance(pp.input, Apply)
        self.assertEquals(lp_scheme, pp.scheme())

    def test_extract_join(self):
        """Extract a join condition from the middle of complex select."""
        s = expression.AND(expression.LTEQ(AttRef("e"), AttRef("f")),
                           expression.AND(
                               expression.EQ(AttRef("c"), AttRef("d")),
                               expression.GT(AttRef("a"), AttRef("b"))))

        lp = StoreTemp('OUTPUT', Select(s, CrossProduct(
            Scan(self.x_key, self.x_scheme),
            Scan(self.y_key, self.y_scheme))))

        self.assertEquals(self.get_num_select_conjuncs(lp), 3)

        pp = self.logical_to_physical(lp)

        # non-equijoin conditions should get pushed separately below the join
        self.assertIsInstance(pp.input, Join)
        self.assertEquals(self.get_count(pp, CrossProduct), 0)
        self.assertEquals(self.get_count(pp, Select), 2)

        self.db.evaluate(pp)
        result = self.db.get_temp_table('OUTPUT')
        self.assertEquals(result, self.expected)

    def test_multi_condition_join(self):
        s = expression.AND(expression.EQ(AttRef("c"), AttRef("d")),
                           expression.EQ(AttRef("a"), AttRef("f")))

        lp = StoreTemp('OUTPUT', Select(s, CrossProduct(
            Scan(self.x_key, self.x_scheme),
            Scan(self.y_key, self.y_scheme))))

        self.assertEquals(self.get_num_select_conjuncs(lp), 2)

        pp = self.logical_to_physical(lp)
        self.assertEquals(self.get_count(pp, CrossProduct), 0)
        self.assertEquals(self.get_count(pp, Select), 0)

        expected = collections.Counter(
            [(a, b, c, d, e, f) for (a, b, c) in self.x_data
             for (d, e, f) in self.y_data if a == f and c == d])

        self.db.evaluate(pp)
        result = self.db.get_temp_table('OUTPUT')
        self.assertEquals(result, expected)

    def test_multiway_join_left_deep(self):

        query = """
        T = SCAN(public:adhoc:Z);
        U = [FROM T AS T1, T AS T2, T AS T3
             WHERE T1.dst==T2.src AND T2.dst==T3.src
             EMIT T1.src AS x, T3.dst AS y];
        STORE(U, OUTPUT);
        """

        lp = self.get_logical_plan(query)
        self.assertEquals(self.get_count(lp, CrossProduct), 2)
        self.assertEquals(self.get_count(lp, Join), 0)

        pp = self.logical_to_physical(lp)
        self.assertEquals(self.get_count(pp, CrossProduct), 0)
        self.assertEquals(self.get_count(pp, Join), 2)
        self.assertEquals(self.get_count(pp, MyriaShuffleProducer), 4)
        self.assertEquals(self.get_count(pp, NaryJoin), 0)
        self.assertEquals(self.get_count(pp, MyriaHyperShuffleProducer), 0)

        self.db.evaluate(pp)
        result = self.db.get_table('OUTPUT')
        self.assertEquals(result, self.expected2)

    def test_multiway_join_hyper_cube(self):

        query = """
        T = SCAN(public:adhoc:Z);
        U = [FROM T AS T1, T AS T2, T AS T3
             WHERE T1.dst==T2.src AND T2.dst==T3.src
             EMIT T1.src AS x, T3.dst AS y];
        STORE(U, OUTPUT);
        """

        lp = self.get_logical_plan(query)
        self.assertEquals(self.get_count(lp, CrossProduct), 2)
        self.assertEquals(self.get_count(lp, Join), 0)

        pp = self.logical_to_physical(lp, hypercube=True)
        self.assertEquals(self.get_count(pp, CrossProduct), 0)
        self.assertEquals(self.get_count(pp, Join), 0)
        self.assertEquals(self.get_count(pp, MyriaShuffleProducer), 0)
        self.assertEquals(self.get_count(pp, NaryJoin), 1)
        self.assertEquals(self.get_count(pp, MyriaHyperShuffleProducer), 3)

        self.db.evaluate(pp)
        result = self.db.get_table('OUTPUT')
        self.assertEquals(result, self.expected2)

    def test_naryjoin_merge(self):
        query = """
        T1 = scan(public:adhoc:Z);
        T2 = [from T1 emit count(dst) as dst, src];
        T3 = scan(public:adhoc:Z);
        twohop = [from T1, T2, T3
                  where T1.dst = T2.src and T2.dst = T3.src
                  emit *];
        store(twohop, anothertwohop);
        """
        statements = self.parser.parse(query)
        self.processor.evaluate(statements)
        lp = self.processor.get_logical_plan()
        pp = self.logical_to_physical(lp, hypercube=True)
        self.assertEquals(self.get_count(pp, NaryJoin), 0)

    def test_right_deep_join(self):
        """Test pushing a selection into a right-deep join tree.

        Myrial doesn't emit these, so we need to cook up a plan by hand."""

        s = expression.AND(expression.EQ(AttIndex(1), AttIndex(2)),
                           expression.EQ(AttIndex(3), AttIndex(4)))

        lp = Apply([('x', AttIndex(0)), ('y', AttIndex(5))],
                   Select(s,
                          CrossProduct(Scan(self.z_key, self.z_scheme),
                                       CrossProduct(
                                           Scan(self.z_key, self.z_scheme),
                                           Scan(self.z_key, self.z_scheme)))))
        lp = StoreTemp('OUTPUT', lp)

        self.assertEquals(self.get_count(lp, CrossProduct), 2)

        pp = self.logical_to_physical(lp)
        self.assertEquals(self.get_count(pp, CrossProduct), 0)

        self.db.evaluate(pp)

        result = self.db.get_temp_table('OUTPUT')
        self.assertEquals(result, self.expected2)

    def test_explicit_shuffle(self):
        """Test of a user-directed partition operation."""

        query = """
        T = SCAN(public:adhoc:X);
        STORE(T, OUTPUT, [$2, b]);
        """
        statements = self.parser.parse(query)
        self.processor.evaluate(statements)
        lp = self.processor.get_logical_plan()

        self.assertEquals(self.get_count(lp, Shuffle), 1)

        for op in lp.walk():
            if isinstance(op, Shuffle):
                self.assertEquals(op.columnlist, [AttIndex(2), AttIndex(1)])

    def test_shuffle_before_distinct(self):
        query = """
        T = DISTINCT(SCAN(public:adhoc:Z));
        STORE(T, OUTPUT);
        """

        pp = self.get_physical_plan(query)
        self.assertEquals(self.get_count(pp, Distinct), 2)  # distributed
        first = True
        for op in pp.walk():
            if isinstance(op, Distinct):
                self.assertIsInstance(op.input, MyriaShuffleConsumer)
                self.assertIsInstance(op.input.input, MyriaShuffleProducer)
                break

    def test_shuffle_before_difference(self):
        query = """
        T = DIFF(SCAN(public:adhoc:Z), SCAN(public:adhoc:Z));
        STORE(T, OUTPUT);
        """

        pp = self.get_physical_plan(query)
        self.assertEquals(self.get_count(pp, Difference), 1)
        for op in pp.walk():
            if isinstance(op, Difference):
                self.assertIsInstance(op.left, MyriaShuffleConsumer)
                self.assertIsInstance(op.left.input, MyriaShuffleProducer)
                self.assertIsInstance(op.right, MyriaShuffleConsumer)
                self.assertIsInstance(op.right.input, MyriaShuffleProducer)

    def test_bug_240_broken_remove_unused_columns_rule(self):
        query = """
        particles = empty(nowGroup:int, timestep:int, grp:int);

        haloTable1 = [from particles as P
                      emit P.nowGroup,
                           (P.timestep+P.grp) as halo,
                           count(*) as totalParticleCount];

        haloTable2 = [from haloTable1 as H, particles as P
                      where H.nowGroup = P.nowGroup
                      emit *];
        store(haloTable2, OutputTemp);
        """

        # This is it -- just test that we can get the physical plan and
        # compile to JSON. See https://github.com/uwescience/raco/issues/240
        pp = self.execute_query(query, output='OutputTemp')

    def test_broadcast_cardinality_right(self):
        # x and y have the same cardinality, z is smaller
        query = """
        x = scan({x});
        y = scan({y});
        z = scan({z});
        out = [from x, z emit *];
        store(out, OUTPUT);
        """.format(x=self.x_key, y=self.y_key, z=self.z_key)

        pp = self.get_physical_plan(query)
        counter = 0
        for op in pp.walk():
            if isinstance(op, CrossProduct):
                counter += 1
                self.assertIsInstance(op.right, MyriaBroadcastConsumer)
        self.assertEquals(counter, 1)

    def test_broadcast_cardinality_left(self):
        # x and y have the same cardinality, z is smaller
        query = """
        x = scan({x});
        y = scan({y});
        z = scan({z});
        out = [from z, y emit *];
        store(out, OUTPUT);
        """.format(x=self.x_key, y=self.y_key, z=self.z_key)

        pp = self.get_physical_plan(query)
        counter = 0
        for op in pp.walk():
            if isinstance(op, CrossProduct):
                counter += 1
                self.assertIsInstance(op.left, MyriaBroadcastConsumer)
        self.assertEquals(counter, 1)

    def test_broadcast_cardinality_with_agg(self):
        # x and y have the same cardinality, z is smaller
        query = """
        x = scan({x});
        y = countall(scan({y}));
        z = scan({z});
        out = [from y, z emit *];
        store(out, OUTPUT);
        """.format(x=self.x_key, y=self.y_key, z=self.z_key)

        pp = self.get_physical_plan(query)
        counter = 0
        for op in pp.walk():
            if isinstance(op, CrossProduct):
                counter += 1
                self.assertIsInstance(op.left, MyriaBroadcastConsumer)
        self.assertEquals(counter, 1)

    def test_relation_cardinality(self):
        query = """
        x = scan({x});
        out = [from x as x1, x as x2 emit *];
        store(out, OUTPUT);
        """.format(x=self.x_key)
        lp = self.get_logical_plan(query)
        self.assertIsInstance(lp, Sequence)
        self.assertEquals(1, len(lp.children()))
        self.assertEquals(sum(self.x_data.values()) ** 2,
                          lp.children()[0].num_tuples())

    def test_relation_physical_cardinality(self):
        query = """
        x = scan({x});
        out = [from x as x1, x as x2 emit *];
        store(out, OUTPUT);
        """.format(x=self.x_key)

        pp = self.get_physical_plan(query)
        self.assertEquals(sum(self.x_data.values()) ** 2,
                          pp.num_tuples())

    def test_catalog_cardinality(self):
        self.assertEquals(sum(self.x_data.values()),
                          self.db.num_tuples(self.x_key))
        self.assertEquals(sum(self.y_data.values()),
                          self.db.num_tuples(self.y_key))
        self.assertEquals(sum(self.z_data.values()),
                          self.db.num_tuples(self.z_key))

    def test_groupby_to_distinct(self):
        query = """
        x = scan({x});
        y = select $0, count(*) from x;
        z = select $0 from y;
        store(z, OUTPUT);
        """.format(x=self.x_key)

        lp = self.get_logical_plan(query)
        self.assertEquals(self.get_count(lp, GroupBy), 1)
        self.assertEquals(self.get_count(lp, Distinct), 0)

        pp = self.logical_to_physical(copy.deepcopy(lp))
        self.assertEquals(self.get_count(pp, GroupBy), 0)
        self.assertEquals(self.get_count(pp, Distinct), 2)  # distributed

        self.assertEquals(self.db.evaluate(lp), self.db.evaluate(pp))

    def test_groupby_to_lesser_groupby(self):
        query = """
        x = scan({x});
        y = select $0, count(*), sum($1) from x;
        z = select $0, $2 from y;
        store(z, OUTPUT);
        """.format(x=self.x_key)

        lp = self.get_logical_plan(query)
        self.assertEquals(self.get_count(lp, GroupBy), 1)
        for op in lp.walk():
            if isinstance(op, GroupBy):
                self.assertEquals(len(op.grouping_list), 1)
                self.assertEquals(len(op.aggregate_list), 2)

        pp = self.logical_to_physical(copy.deepcopy(lp))
        self.assertEquals(self.get_count(pp, GroupBy), 2)  # distributed
        for op in pp.walk():
            if isinstance(op, GroupBy):
                self.assertEquals(len(op.grouping_list), 1)
                self.assertEquals(len(op.aggregate_list), 1)

        self.assertEquals(self.db.evaluate(lp), self.db.evaluate(pp))

    def __run_uda_test(self, uda_state=None):
        scan = Scan(self.x_key, self.x_scheme)

        init_ex = expression.NumericLiteral(0)
        update_ex = expression.PLUS(expression.NamedStateAttributeRef("value"),
                                    AttIndex(1))
        emit_ex = expression.UdaAggregateExpression(
            expression.NamedStateAttributeRef("value"), uda_state)
        statemods = [StateVar("value", init_ex, update_ex)]

        log_gb = GroupBy([AttIndex(0)], [emit_ex], scan, statemods)

        lp = StoreTemp('OUTPUT', log_gb)
        pp = self.logical_to_physical(copy.deepcopy(lp))

        self.db.evaluate(lp)
        log_result = self.db.get_temp_table('OUTPUT')

        self.db.delete_temp_table('OUTPUT')
        self.db.evaluate(pp)
        phys_result = self.db.get_temp_table('OUTPUT')

        self.assertEquals(log_result, phys_result)
        self.assertEquals(len(log_result), 15)

        self.assertEquals(self.get_count(pp, MyriaShuffleProducer), 1)
        self.assertEquals(self.get_count(pp, MyriaShuffleConsumer), 1)

        return pp

    def test_non_decomposable_uda(self):
        """Test that optimization preserves the value of a non-decomposable UDA
        """
        pp = self.__run_uda_test()

        for op in pp.walk():
            if isinstance(op, MyriaShuffleProducer):
                self.assertEquals(op.hash_columns, [AttIndex(0)])
                self.assertEquals(self.get_count(op, GroupBy), 0)

    def test_decomposable_uda(self):
        """Test that optimization preserves the value of decomposable UDAs"""
        lemits = [expression.UdaAggregateExpression(
                  expression.NamedStateAttributeRef("value"))]
        remits = copy.deepcopy(lemits)

        init_ex = expression.NumericLiteral(0)
        update_ex = expression.PLUS(expression.NamedStateAttributeRef("value"),
                                    AttIndex(1))
        lstatemods = [StateVar("value", init_ex, update_ex)]
        rstatemods = copy.deepcopy(lstatemods)

        uda_state = expression.DecomposableAggregateState(
            lemits, lstatemods, remits, rstatemods)
        pp = self.__run_uda_test(uda_state)

        self.assertEquals(self.get_count(pp, GroupBy), 2)

        for op in pp.walk():
            if isinstance(op, MyriaShuffleProducer):
                self.assertEquals(op.hash_columns, [AttIndex(0)])
                self.assertEquals(self.get_count(op, GroupBy), 1)

    def test_successful_append(self):
        """Insert an append if storing a relation into itself with a
        UnionAll."""
        query = """
        x = scan({x});
        y = select $0 from x;
        y2 = select $1 from x;
        y = y+y2;
        store(y, OUTPUT);
        """.format(x=self.x_key)

        lp = self.get_logical_plan(query, apply_chaining=False)
        self.assertEquals(self.get_count(lp, ScanTemp), 5)
        self.assertEquals(self.get_count(lp, StoreTemp), 4)
        self.assertEquals(self.get_count(lp, AppendTemp), 0)
        self.assertEquals(self.get_count(lp, Store), 1)
        self.assertEquals(self.get_count(lp, Scan), 1)

        pp = self.logical_to_physical(copy.deepcopy(lp))
        self.assertEquals(self.get_count(pp, ScanTemp), 4)
        self.assertEquals(self.get_count(pp, StoreTemp), 3)
        self.assertEquals(self.get_count(pp, AppendTemp), 1)
        self.assertEquals(self.get_count(pp, Store), 1)
        self.assertEquals(self.get_count(pp, Scan), 1)

        self.assertEquals(self.db.evaluate(lp), self.db.evaluate(pp))

    def test_failed_append(self):
        """Do not insert an append when the tuples to be appended
        depend on the relation itself."""

        # NB test in both the left and right directions
        # left: y = y + y2
        # right: y = y2 + y
        query = """
        x = scan({x});
        y = select $0, $1 from x;
        t = empty(a:int);
        y2 = select $1, $1 from y;
        y = y+y2;
        t = empty(a:int);
        y3 = select $1, $1 from y;
        y = y3+y;
        s = empty(a:int);
        store(y, OUTPUT);
        """.format(x=self.x_key)

        lp = self.get_logical_plan(query, dead_code_elimination=False)
        self.assertEquals(self.get_count(lp, AppendTemp), 0)

        # No AppendTemp
        pp = self.logical_to_physical(copy.deepcopy(lp))
        self.assertEquals(self.get_count(pp, AppendTemp), 0)

        self.assertEquals(self.db.evaluate(lp), self.db.evaluate(pp))

    def test_push_work_into_sql(self):
        """Test generation of MyriaQueryScan operator for query with
        projects"""
        query = """
        r3 = scan({x});
        intermediate = select a, c from r3;
        store(intermediate, OUTPUT);
        """.format(x=self.x_key)

        pp = self.get_physical_plan(query, push_sql=True)
        self.assertEquals(self.get_count(pp, Operator), 2)
        self.assertTrue(isinstance(pp.input, MyriaQueryScan))

        expected = collections.Counter([(a, c) for (a, b, c) in self.x_data])

        self.db.evaluate(pp)
        result = self.db.get_table('OUTPUT')
        self.assertEquals(result, expected)

    def test_push_work_into_sql_2(self):
        """Test generation of MyriaQueryScan operator for query with projects
        and a filter"""
        query = """
        r3 = scan({x});
        intermediate = select a, c from r3 where b < 5;
        store(intermediate, OUTPUT);
        """.format(x=self.x_key)

        pp = self.get_physical_plan(query, push_sql=True)
        self.assertEquals(self.get_count(pp, Operator), 2)
        self.assertTrue(isinstance(pp.input, MyriaQueryScan))

        expected = collections.Counter([(a, c)
                                        for (a, b, c) in self.x_data
                                        if b < 5])

        self.db.evaluate(pp)
        result = self.db.get_table('OUTPUT')
        self.assertEquals(result, expected)

    def test_no_push_when_shuffle(self):
        """When data is not co-partitioned, the join should not be pushed."""
        query = """
        r3 = scan({x});
        s3 = scan({y});
        intermediate = select r3.a, s3.f from r3, s3 where r3.b=s3.e;
        store(intermediate, OUTPUT);
        """.format(x=self.x_key, y=self.y_key)

        pp = self.get_physical_plan(query, push_sql=True)
        # Join is not pushed
        self.assertEquals(self.get_count(pp, Join), 1)
        # The projections are pushed into the QueryScan
        self.assertEquals(self.get_count(pp, MyriaQueryScan), 2)
        # We should not need any Apply since there is no rename and no other
        # project.
        self.assertEquals(self.get_count(pp, Apply), 0)

        expected = collections.Counter([(a, f)
                                        for (a, b, c) in self.x_data
                                        for (d, e, f) in self.y_data
                                        if b == e])

        self.db.evaluate(pp)
        result = self.db.get_table('OUTPUT')
        self.assertEquals(result, expected)



"""PageRank unit test.

Example data taken from:
http://select.cs.cmu.edu/code/graphlab/doxygen/html/pagerank_example.html
"""

import collections

import raco.scheme as scheme
import raco.myrial.myrial_test as myrial_test
from raco import types


class PageRankTest(myrial_test.MyrialTestCase):

    edge_table = collections.Counter([
        (0, 3),
        (1, 0),
        (1, 2),
        (2, 0),
        (2, 1),
        (2, 3),
        (3, 0),
        (3, 1),
        (3, 2),
        (3, 4),
        (4, 0),
        (4, 1),
        (4, 2),
        (4, 3),
        (4, 4)])

    edge_schema = scheme.Scheme([("src", types.LONG_TYPE),
                                 ("dst", types.LONG_TYPE)])
    edge_key = "public:adhoc:edges"

    vertex_table = collections.Counter([(x,) for x in range(5)])
    vertex_key = "public:adhoc:vertices"
    vertex_schema = scheme.Scheme([("id", types.LONG_TYPE)])

    def setUp(self):
        super(PageRankTest, self).setUp()

        self.db.ingest(PageRankTest.edge_key,
                       PageRankTest.edge_table,
                       PageRankTest.edge_schema)

        self.db.ingest(PageRankTest.vertex_key,
                       PageRankTest.vertex_table,
                       PageRankTest.vertex_schema)

    def __do_test(self, phile):
        with open(phile) as fh:
            query = fh.read()

        result = self.execute_query(query)
        d = dict(result.elements())

        self.assertAlmostEqual(d[0], 0.23576110832410296)
        self.assertAlmostEqual(d[1], 0.16544845649781043)
        self.assertAlmostEqual(d[2], 0.18370688939571236)
        self.assertAlmostEqual(d[3], 0.3016893082129546)
        self.assertAlmostEqual(d[4], 0.11339423756941983)

    def test_pagerank(self):
        self.__do_test('examples/pagerank.myl')

    def verify_undefined(self, var):
        with self.assertRaises(KeyError):
            self.db.get_temp_table(var)

    def test_pagerank_deadcode(self):
        """Test of page rank with numerous dead code statements."""
        self.__do_test('examples/pagerank_dead.myl')

        # Verify that D0, D1 tables are compiled out
        self.verify_undefined("D0")
        self.verify_undefined("D1")



"""
A library of expressions that can be composed of existing expressions.
"""

from .udf import Function
import raco.expression
from raco.expression import *


def is_defined(function_name):
    return function_name.lower() in EXPRESSIONS


def lookup(function_name, num_args):
    func = EXPRESSIONS.get(function_name.lower())
    if hasattr(func, '__call__'):
        return func(num_args)
    if isinstance(func, dict):
        return func.get(num_args)
    return func


def create_nested_binary(num_args, func):
    if num_args < 2:
        return None

    var = ["x{i}".format(i=i + 1) for i in xrange(num_args)]
    var_refs = [NamedAttributeRef(vstr) for vstr in var]
    return Function(var, reduce(func, var_refs))


# mapping from name -> dict or Function
# the dict is a mapping from arity -> Function
EXPRESSIONS_CASE = {
    'SafeDiv': {
        2: Function(['n', 'd'], Case(
            [(EQ(NamedAttributeRef('d'), NumericLiteral(0)),
              NumericLiteral(0.0))],
            DIVIDE(NamedAttributeRef('n'), NamedAttributeRef('d')))),
        3: Function(['n', 'd', 'default'], Case(
            [(EQ(NamedAttributeRef('d'), NumericLiteral(0)),
              CAST(types.DOUBLE_TYPE, NamedAttributeRef('default')))],
            DIVIDE(NamedAttributeRef('n'), NamedAttributeRef('d'))))
    },
    'TheAnswerToLifeTheUniverseAndEverything': Function(
        [], NumericLiteral(42)),
    'greatest': lambda num_args: create_nested_binary(num_args, GREATER),
    'least': lambda num_args: create_nested_binary(num_args, LESSER),
    'greater': create_nested_binary(2, GREATER),
    'lesser': create_nested_binary(2, LESSER),
    'substr': Function(['str', 'begin', 'end'],
                       SUBSTR([NamedAttributeRef('str'),
                               NamedAttributeRef('begin'),
                               NamedAttributeRef('end')
                               ])),
    'head': Function(['str', 'length'],
                     SUBSTR([NamedAttributeRef('str'),
                             NumericLiteral(0),
                             LESSER(LEN(NamedAttributeRef('str')),
                                    NamedAttributeRef('length'))
                             ])),
    'tail': Function(['str', 'length'],
                     SUBSTR([NamedAttributeRef('str'),
                             GREATER(MINUS(LEN(NamedAttributeRef('str')),
                                           NamedAttributeRef('length')),
                                     NumericLiteral(0)),
                             LEN(NamedAttributeRef('str'))
                             ]))
}


def get_arity(func_class):
    """Return the arity of built-in Myria expressions."""

    if issubclass(func_class, ZeroaryOperator):
        return 0
    elif issubclass(func_class, UnaryOperator):
        return 1
    elif issubclass(func_class, BinaryOperator):
        return 2
    else:
        # Don't handle n-ary functions automatically
        assert False


def one_to_one_function(func_name):
    """Emit a Function object that wraps a Myria built-in expression."""
    func_class = getattr(raco.expression, func_name)
    arity = get_arity(func_class)
    function_args = ['arg%d' % i for i in range(arity)]
    expression_args = [NamedAttributeRef(x) for x in function_args]
    return Function(function_args, func_class(*expression_args))

# Simple functions that map to a single Myria expression; the names here
# must match the corresponding function class in raco.expression.function
ONE_TO_ONE_FUNCS = ['ABS', 'CEIL', 'COS', 'FLOOR', 'LOG', 'SIN', 'SQRT',
                    'TAN', 'LEN', 'POW', 'MAX', 'MIN', 'SUM', 'AVG', 'STDEV',
                    'COUNTALL', 'MD5']

ONE_TO_ONE_EXPRS = {k.lower(): one_to_one_function(k) for k in ONE_TO_ONE_FUNCS}  # noqa

EXPRESSIONS = {k.lower(): v for k, v in EXPRESSIONS_CASE.items()}
EXPRESSIONS.update(ONE_TO_ONE_EXPRS)



import collections
import raco.scheme as scheme
import raco.types as types

"""This class contains fake data used by several unit tests."""


class FakeData(object):
    emp_table = collections.Counter([
        # id dept_id name salary
        (1, 2, "Bill Howe", 25000),
        (2, 1, "Dan Halperin", 90000),
        (3, 1, "Andrew Whitaker", 5000),
        (4, 2, "Shumo Chu", 5000),
        (5, 1, "Victor Almeida", 25000),
        (6, 3, "Dan Suciu", 90000),
        (7, 1, "Magdalena Balazinska", 25000)])

    emp_schema = scheme.Scheme([("id", types.INT_TYPE),
                                ("dept_id", types.INT_TYPE),
                                ("name", types.STRING_TYPE),
                                ("salary", types.LONG_TYPE)])

    emp_key = "public:adhoc:employee"

    dept_table = collections.Counter([
        (1, "accounting", 5),
        (2, "human resources", 2),
        (3, "engineering", 2),
        (4, "sales", 7)])

    dept_schema = scheme.Scheme([("id", types.LONG_TYPE),
                                 ("name", types.STRING_TYPE),
                                 ("manager", types.LONG_TYPE)])

    dept_key = "public:adhoc:department"

    numbers_table = collections.Counter([
        (1, 3),
        (2, 5),
        (3, -2),
        (16, -4.3)])

    numbers_schema = scheme.Scheme([("id", types.LONG_TYPE),
                                    ("val", types.DOUBLE_TYPE)])

    numbers_key = "public:adhoc:numbers"



from collections import Counter, defaultdict
from raco.language.sql.test_case import SQLTestCase

import raco.scheme as scheme
import raco.types as types


class TestScheme(SQLTestCase):
    """Test that we can convert Raco schemes to SQL Alchemy schemes"""
    def test_simple_scheme(self):
        sch = scheme.Scheme()
        sch.addAttribute('w', types.FLOAT_TYPE)
        sch.addAttribute('x', types.INT_TYPE)
        sch.addAttribute('y', types.LONG_TYPE)
        sch.addAttribute('z', types.STRING_TYPE)

        self.db.add_table('simple', sch)

        sch2 = self.db.get_scheme('simple')
        self.assertEquals(sch, sch2)


class TestQuery(SQLTestCase):
    """Test actually compiling plans to SQL."""
    def test_catalog(self):
        self.assertEquals(len(self.emp_table),
                          self.db.num_tuples(self.emp_key))

    def test_simple_scan(self):
        query = """x = scan({emp});
        store(x, OUTPUT);""".format(emp=self.emp_key)

        expected = Counter(self.emp_table)
        self.execute(query, expected)

    def test_column_select(self):
        query = """x = scan({emp});
        y = [from x emit $2];
        store(y, OUTPUT);""".format(emp=self.emp_key)

        expected = Counter((x[2],) for x in self.emp_table)
        self.execute(query, expected)

    def test_rename_column_select(self):
        query = """x = scan({emp});
        y = [from x emit $0 as a, $0 as b];
        store(y, OUTPUT);""".format(emp=self.emp_key)

        expected = Counter((x[0], x[0]) for x in self.emp_table)
        self.execute(query, expected)

    def test_count_query(self):
        query = """x = countall(scan({emp}));
        store(x, OUTPUT);""".format(emp=self.emp_key)

        expected = Counter([(len(self.emp_table),)])
        self.execute(query, expected)

    def test_theta_join_query(self):
        query = """x = scan({emp});
        y = scan({emp});
        z = [from x,y where x.salary < y.salary emit x.*];
        store(z, OUTPUT);""".format(emp=self.emp_key)

        expected = Counter((a, b, c, d)
                           for (a, b, c, d) in self.emp_table
                           for (_, _, _, d2) in self.emp_table
                           if d < d2)
        self.execute(query, expected)

    def test_join_query(self):
        query = """x = scan({emp});
        y = scan({emp});
        z = [from x,y where x.salary = y.salary emit count(*) as cnt];
        store(z, OUTPUT);""".format(emp=self.emp_key)

        size = len([1
                    for (_, _, _, d) in self.emp_table
                    for (_, _, _, d2) in self.emp_table
                    if d == d2])
        expected = Counter([(size,)])
        self.execute(query, expected)

    def test_complex_agg_query(self):
        query = """x = scan({emp});
        z = [from x emit dept_id, max(salary) as max_salary];
        store(z, OUTPUT);""".format(emp=self.emp_key)

        d = defaultdict(int)
        for _, dept_id, _, salary in self.emp_table:
            d[dept_id] = max(d[dept_id], salary)
        expected = Counter(d.items())
        self.execute(query, expected)




import collections

import raco.algebra
import raco.scheme as scheme
import raco.myrial.myrial_test as myrial_test
from raco import types


class ReachableTest(myrial_test.MyrialTestCase):

    edge_table = collections.Counter([
        (1, 2),
        (2, 3),
        (3, 4),
        (4, 3),
        (3, 5),
        (4, 13),
        (5, 4),
        (1, 9),
        (7, 1),
        (6, 1),
        (10, 11),
        (11, 12),
        (12, 10),
        (13, 4),
        (10, 1)])

    edge_schema = scheme.Scheme([("src", types.LONG_TYPE),
                                 ("dst", types.LONG_TYPE)])
    edge_key = "public:adhoc:edges"

    def setUp(self):
        super(ReachableTest, self).setUp()

        self.db.ingest(ReachableTest.edge_key,
                       ReachableTest.edge_table,
                       ReachableTest.edge_schema)

    def test_reachable(self):
        with open('examples/reachable.myl') as fh:
            query = fh.read()

        expected = collections.Counter([
            (1,),
            (2,),
            (3,),
            (4,),
            (5,),
            (9,),
            (13,),
        ])

        self.check_result(query, expected, skip_json=True)

    def test_multi_condition_join(self):
        query = """
        Edge = SCAN(public:adhoc:edges);
        Symmetric = [FROM Edge AS E1, Edge AS E2
                     WHERE E1.src==E2.dst
                       AND E2.src==E1.dst
                       AND E1.src < E1.dst
                     EMIT E1.src AS src, E1.dst AS dst];
        STORE(Symmetric, OUTPUT);
        """
        table = ReachableTest.edge_table
        expected = collections.Counter(
            [(a, b) for (a, b) in table for (c, d) in table
             if a == d and b == c and a < b])
        self.check_result(query, expected)

    def test_cross_plus_selection_becomes_join(self):
        """Test that the optimizer compiles away cross-products."""
        with open('examples/reachable.myl') as fh:
            query = fh.read()

        def plan_contains_cross(plan):
            def f(op):
                if isinstance(op, raco.algebra.CrossProduct) and not \
                   isinstance(op.left, raco.algebra.SingletonRelation):
                    yield True

            return any(plan.postorder(f))

        statements = self.parser.parse(query)
        self.processor.evaluate(statements)

        lp = self.processor.get_logical_plan()
        self.assertTrue(plan_contains_cross(lp))

        pp = self.processor.get_physical_plan()
        self.assertFalse(plan_contains_cross(pp))



from raco import algebra
import raco.language as language
from pipelines import Pipelined
from raco.utility import emit
import raco.viz as viz
import os
from raco.utility import colored

import logging
LOG = logging.getLogger(__name__)


"""
Apply rules to an expression
If successful, output will
only involve operators in the
target algebra
"""


class PlanWriter():

    def __init__(self, template="wip-%02d.physical.dot", limit=20):
        self.ind = 0
        self.template = template
        self.limit = limit
        self.enabled = os.environ.get('RACO_OPTIMIZER_GRAPHS') in \
            ['true', 'True', 't', 'T', '1', 'yes', 'y']

    def write_if_enabled(self, plan, title):
        if self.enabled:
            with open(self.template % self.ind, 'w') as dwf:
                dwf.write(viz.operator_to_dot(plan, title=title))

        self.ind += 1


def optimize_by_rules(expr, rules):
    writer = PlanWriter()
    writer.write_if_enabled(expr, "before rules")

    for rule in rules:
        def recursiverule(e):
            newe = rule(e)
            writer.write_if_enabled(newe, str(rule))

            LOG.debug("apply rule %s\n" +
                      colored("  -", "red") + " %s" + "\n" +
                      colored("  +", "green") + " %s", rule, e, newe)
            newe.apply(recursiverule)

            return newe
        expr = recursiverule(expr)

    return expr


def optimize(expr, target, **kwargs):
    """Fire the rule-based optimizer on an expression.  Fire all rules in the
    target algebra."""
    assert isinstance(expr, algebra.Operator)
    assert isinstance(target, language.Algebra), type(target)

    return optimize_by_rules(expr, target.opt_rules(**kwargs))


def compile(expr):
    """Compile physical plan to linearized form for execution"""
    # TODO: Fix this
    algebra.reset()
    exprcode = []

    # TODO, actually use Parallel[Store...]]? Right now assumes it
    if isinstance(expr, (algebra.Sequence, algebra.Parallel)):
        assert len(expr.children()) == 1, "expected single expression only"
        store_expr = expr.children()[0]
    else:
        store_expr = expr

    assert isinstance(store_expr, algebra.Store)
    assert len(store_expr.children()) == 1, "expected single expression only"

    lang = store_expr.language()

    if isinstance(store_expr, Pipelined):
        body = lang.body(store_expr.compilePipeline())
    else:
        body = lang.body(expr)

    exprcode.append(emit(body))
    return emit(*exprcode)